# -*- coding: utf-8 -*-

from plone import api
from plone.app.contenttypes.migration.dxmigration import ContentMigrator
from plone.app.contenttypes.migration.migration import migrate as pac_migrate
from Products.CMFPlone.interfaces.constrains import IConstrainTypes
from Products.CMFPlone.utils import base_hasattr
from Products.PloneMeeting.migrations import logger
from Products.PloneMeeting.migrations import Migrator
from Products.ZCatalog.ProgressHandler import ZLogHandler
from zope.i18n import translate


class MeetingCategoryMigrator(ContentMigrator):
    """For DashboardPODTemplates created after imio.dashboard 0.28 where
       meta_type was removed and so 'Dexterity Item' by default."""
    src_portal_type = 'MeetingCategory'
    src_meta_type = 'MeetingCategory'
    dst_portal_type = 'meetingcategory'
    dst_meta_type = None  # not used

    def migrate_atctmetadata(self):
        """Override to not migrate exclude_from_nav because it does not exist by default
           and it takes parent's value that is an instancemethod and fails at transaction commit..."""
        pass

    def migrate_schema_fields(self):
        self.new.category_id = self.old.getCategoryId()
        self.new.using_groups = self.old.getUsingGroups()
        self.new.category_mapping_when_cloning_to_other_mc = self.old.getCategoryMappingsWhenCloningToOtherMC()
        self.new.groups_in_charge = self.old.getGroupsInCharge()
        self.new.enabled = api.content.get_state(self.old) == 'active' and True or False
        self.new.reindexObject()


class Migrate_To_4110(Migrator):

    def _migrateMeetingCategoryToDX(self):
        '''Migrate from AT MeetingCategory to DX meetingcategory.'''
        logger.info('Migrating MeetingCategory from AT to DX...')
        # update item classifier
        # migrate references
        pghandler = ZLogHandler(steps=100)
        brains = self.portal.reference_catalog(relationship='ItemClassification')
        pghandler.init('Updating field MeetingItem.classifier...', len(brains))
        pghandler.info('Updating field MeetingItem.classifier...')
        i = 0
        for brain in brains:
            i += 1
            pghandler.report(i)
            relation = brain.getObject()
            item = relation.getSourceObject()
            classifier = relation.getTargetObject()
            item.setClassifier(classifier.getId())
            item.reindexObject(idxs=['getRawClassifier'])
        # deleteReferences in a second phase
        for brain in brains:
            relation = brain.getObject()
            item = relation.getSourceObject()
            item.deleteReferences('ItemClassification')
        pghandler.finish()

        logger.info('Migrating categories and classifiers in configuration...')
        # make sure new portal_type meetingcategory is installed
        self.ps.runImportStepFromProfile('profile-Products.PloneMeeting:default', 'typeinfo')
        # make sure no workflow used for meetingcategory
        self.wfTool.setChainForPortalTypes(('meetingcategory', ), ('', ))
        # adapt allowed_types for each MeetingConfig.categories/classifiers folders
        for cfg in self.tool.objectValues('MeetingConfig'):
            for folder_id in ('categories', 'classifiers'):
                constrain = IConstrainTypes(getattr(cfg, folder_id))
                constrain.setConstrainTypesMode(1)
                allowedTypes = ['meetingcategory']
                constrain.setLocallyAllowedTypes(allowedTypes)
                constrain.setImmediatelyAddableTypes(allowedTypes)
        # migrate to DX
        pac_migrate(self.portal, MeetingCategoryMigrator)
        self.removeUnusedPortalTypes(portal_types=['MeetingCategory'])
        # add meetingcategory to types_not_searched
        props = api.portal.get_tool('portal_properties').site_properties
        nsTypes = props.getProperty('types_not_searched')
        if 'meetingcategory' not in nsTypes:
            nsTypes = list(nsTypes)
            # MeetingCategory was removed by removeUnusedPortalTypes
            nsTypes.append('meetingcategory')
            props.manage_changeProperties(types_not_searched=tuple(nsTypes))
        logger.info('Done.')

    def _updateOrgsDashboardCollectionColumns(self):
        """Enable column 'PloneGroupUsersGroupsColumn' in contacts collections displaying organizations."""
        logger.info("Updating dashboard organization collections to remove the \"review_state\" column...")
        orgs_searches_folder = self.portal.contacts.get('orgs-searches')
        # enable 'PloneGroupUsersGroupsColumn' just after 'SelectedInPlonegroupColumn'
        for org_coll in orgs_searches_folder.objectValues():
            customViewFields = list(org_coll.customViewFields)
            if 'PloneGroupUsersGroupsColumn' not in customViewFields:
                customViewFields.insert(
                    customViewFields.index('SelectedInPlonegroupColumn') + 1,
                    'PloneGroupUsersGroupsColumn')
                org_coll.customViewFields = customViewFields
        logger.info('Done.')

    def _enableDateinputJSResource(self):
        """Make sure '++resource++plone.app.jquerytools.dateinput.js' is enabled in portal_javascripts."""
        logger.info("Enabling '++resource++plone.app.jquerytools.dateinput.js' in portal_javascripts...")
        resource = self.portal.portal_javascripts.getResource('++resource++plone.app.jquerytools.dateinput.js')
        resource.setEnabled(True)
        logger.info('Done.')

    def _addPageBreakStyleToCKEditor(self):
        """Add new style 'page-break' to CKeditor styles."""
        logger.info("Adding style 'page-break' to CKEditor styles...")
        cke_props = self.portal.portal_properties.ckeditor_properties
        if cke_props.menuStyles.find('page-break') == -1:
            msg_page_break = translate('ckeditor_style_page_break',
                                       domain='PloneMeeting',
                                       context=self.request).encode('utf-8')
            menuStyles = cke_props.menuStyles
            page_break_style = "{{ name : '{0}'\t\t, element : 'p', attributes : " \
                "{{ 'class' : 'page-break' }} }},\n]".format(msg_page_break)
            # last element, check if we need a ',' before or not...
            strippedMenuStyles = menuStyles.replace(' ', '').replace('\n', '').replace('\r', '')
            if ',]' not in strippedMenuStyles:
                menuStyles = menuStyles.replace('\n]', ']')
                page_break_style = ",\n" + page_break_style
            menuStyles = menuStyles.replace(']', page_break_style)
            cke_props.menuStyles = menuStyles
            logger.info("Style 'page-break' was added...")
        else:
            logger.info("Style 'page-break' already exists and was not added...")
        logger.info('Done.')

    def _migrateToMeetingItemTemplatesToStoreAsAnnex(self):
        """Single value field MeetingConfig.meetingItemTemplateToStoreAsAnnex was moved
           to multi valued field MeetingConfig.meetingItemTemplatesToStoreAsAnnex."""
        logger.info("Migrating MeetingConfig.meetingItemTemplateToStoreAsAnnex to "
                    "MeetingConfig.meetingItemTemplatesToStoreAsAnnex...")
        old_field_name = 'meetingItemTemplateToStoreAsAnnex'
        new_field_name = 'meetingItemTemplatesToStoreAsAnnex'
        for cfg in self.tool.objectValues('MeetingConfig'):
            if base_hasattr(cfg, old_field_name):
                old_field_value = getattr(cfg, old_field_name)
                if old_field_value:
                    setattr(cfg, new_field_name, [old_field_value])
        logger.info('Done.')

    def run(self, from_migration_to_41=False):
        logger.info('Migrating to PloneMeeting 4110...')
        self._migrateMeetingCategoryToDX()
        self._updateOrgsDashboardCollectionColumns()
        self._enableDateinputJSResource()
        self._addPageBreakStyleToCKEditor()
        self._migrateToMeetingItemTemplatesToStoreAsAnnex()
        # update collective.contact.plonegroup
        self.upgradeAll(omit=['Products.PloneMeeting:default',
                              self.profile_name.replace('profile-', '')])


def migrate(context):
    '''This migration function will:

       1) Migrate MeetingCategory to meetingcategory.
       2) Enable column 'PloneGroupUsersGroupsColumn' of for contacts collections displaying organizations;
       3) Make sure '++resource++plone.app.jquerytools.dateinput.js' is enabled in portal_javascripts;
       4) Add new style 'page-break' to CKEditor;
       5) Migrate MeetingConfig.meetingItemTemplateToStoreAsAnnex to MeetingConfig.meetingItemTemplatesToStoreAsAnnex;
       6) Apply every pending upgrades.
    '''
    migrator = Migrate_To_4110(context)
    migrator.run()
    migrator.finish()



# -*- coding: utf-8 -*-
#
# File: testPortlets.py
#
# GNU General Public License (GPL)
#

from imio.helpers.cache import cleanRamCacheFor
from plone.portlets.interfaces import IPortletManager
from plone.portlets.interfaces import IPortletRenderer
from Products.Archetypes.event import ObjectEditedEvent
from Products.PloneMeeting.browser import portlet_plonemeeting
from Products.PloneMeeting.browser import portlet_todo
from Products.PloneMeeting.config import ITEM_DEFAULT_TEMPLATE_ID
from Products.PloneMeeting.tests.PloneMeetingTestCase import PloneMeetingTestCase
from zope.component import getMultiAdapter
from zope.component import getUtility
from zope.event import notify
from zope.lifecycleevent import ObjectModifiedEvent


class testPortlets(PloneMeetingTestCase):
    '''Tests the MeetingItem class methods.'''

    def _setup_portlets(self):
        """ """
        self.changeUser('pmCreator1')
        searches_items = self.getMeetingFolder(self.meetingConfig).searches_items
        self.view = self.portal.restrictedTraverse('@@plone')
        self.manager = getUtility(IPortletManager,
                                  name='plone.leftcolumn',
                                  context=self.portal)
        self.portlet_pm_assignment = portlet_plonemeeting.Assignment()
        self.portlet_pm_renderer = getMultiAdapter(
            (searches_items,
             self.request,
             self.view,
             self.manager,
             self.portlet_pm_assignment),
            IPortletRenderer)
        self.portlet_todo_assignment = portlet_todo.Assignment(batch_size=5,
                                                               title_length=100)
        self.portlet_todo_renderer = getMultiAdapter(
            (searches_items,
             self.request,
             self.view,
             self.manager,
             self.portlet_todo_assignment),
            IPortletRenderer)

    def test_pm_PortletPMAvailableTemplates(self):
        '''Test the portlet_plonemeeting itemTemplates icon that is shown if item templates
           are defined in the configuration, no matter current user have item templates or not,
           this way we have something coherent between users, even for users without itemTemplates.'''
        cfg = self.meetingConfig
        # remove every templates and add one restricted to 'developers'
        self.changeUser('siteadmin')
        self._removeConfigObjectsFor(cfg, folders=['recurringitems', 'itemtemplates'])
        itemTemplate = self.create('MeetingItemTemplate')
        itemTemplate.setTemplateUsingGroups(('developers', ))
        itemTemplate.reindexObject(idxs=['templateUsingGroups', ])
        # pmCreator1 is member of 'developers'
        self.changeUser('pmCreator1')
        pmFolder1 = self.getMeetingFolder()
        itemsCategory = pmFolder1.restrictedTraverse('@@render_collection_widget_category')
        itemsCategory(widget=None)
        self.assertTrue(itemsCategory.hasTemplateItems())
        # pmCreator2 is member of 'vendors'
        self.changeUser('pmCreator2')
        pmFolder2 = self.getMeetingFolder()
        itemsCategory = pmFolder2.restrictedTraverse('@@render_collection_widget_category')
        itemsCategory(widget=None)
        # clean ram.cache even if cache is still correct because the same for every users
        notify(ObjectEditedEvent(cfg))
        self.assertTrue(itemsCategory.hasTemplateItems())
        # no matter actually there are no itemTemplates available for him...
        self.assertFalse(cfg.getItemTemplates(as_brains=True, onlyActive=True, filtered=True))

    def test_pm_PortletPMRestrictCreateItemFromEmptyTemplate(self):
        '''Test that the "Crate item from empty template" may be restricted
           to some groups, this let's make the link "Create empty item" available
           to only some groups.'''
        cfg = self.meetingConfig
        empty_item_template = cfg.itemtemplates.get(ITEM_DEFAULT_TEMPLATE_ID)
        empty_item_template_uid = empty_item_template.UID()
        # by default template is not restricted so available to everybody
        self.changeUser('pmCreator1')
        pmFolder1 = self.getMeetingFolder()
        itemsCategory = pmFolder1.restrictedTraverse('@@render_collection_widget_category')
        itemsCategory(widget=None)
        self.assertEqual(itemsCategory._get_default_item_template_UID(), empty_item_template_uid)
        self.changeUser('pmCreator2')
        pmFolder2 = self.getMeetingFolder()
        itemsCategory = pmFolder2.restrictedTraverse('@@render_collection_widget_category')
        itemsCategory(widget=None)
        self.assertEqual(itemsCategory._get_default_item_template_UID(), empty_item_template_uid)
        # restrict it to developers
        self.changeUser('siteadmin')
        empty_item_template.setTemplateUsingGroups([self.developers_uid])
        empty_item_template.reindexObject(idxs=['templateUsingGroups', ])
        # available for pmCreator1 that is member of developers
        self.changeUser('pmCreator1')
        itemsCategory = pmFolder1.restrictedTraverse('@@render_collection_widget_category')
        itemsCategory(widget=None)
        self.assertEqual(itemsCategory._get_default_item_template_UID(), empty_item_template_uid)
        # not available for pmCreator2 that is not member of developers
        self.changeUser('pmCreator2')
        itemsCategory = pmFolder2.restrictedTraverse('@@render_collection_widget_category')
        itemsCategory(widget=None)
        self.assertIsNone(itemsCategory._get_default_item_template_UID())

    def test_pm_FromPortletTodo(self):
        """While getting searches in portlet_todo, the TAL condition for searches have
           a 'fromPortletTodo=True', it is not the case in the portlet_plonemeeting, this way
           we may know that we are in portlet_todo or portlet_plonemeeting and display
           searches using a different condition."""
        cfg = self.meetingConfig
        self._setup_portlets()
        # by default, no condition, viewable in both portlets
        searches = cfg.searches
        searchAllItems = searches.searches_items.searchallitems
        searchAllItems.tal_condition = ''
        searchAllItemsUID = searchAllItems.UID()
        # select 'searchallitems' in the MeetingConfig.toDoListSearches
        cfg.setToDoListSearches([searchAllItemsUID])

        # viewable in portlet_plonemeeting
        self.changeUser('pmCreator1')
        item = self.create('MeetingItem')
        item_url = item.absolute_url()
        cleanRamCacheFor('Products.PloneMeeting.adapters.compute_criteria')
        self.assertTrue(searchAllItemsUID in self.portlet_pm_renderer.render())
        # and viewable in portlet_todo
        self.request.set('load_portlet_todo', True)
        self.assertTrue(searchAllItemsUID in self.portlet_todo_renderer.render())
        self.assertTrue(item_url in self.portlet_todo_renderer.render())

        # set 'python: fromPortletTodo' as condition for a search, it will be displayed
        # in the portlet_todo but not in the portlet_plonemeeting
        searchAllItems.tal_condition = 'python: fromPortletTodo'
        notify(ObjectModifiedEvent(searchAllItems))
        self._setup_portlets()
        # not viewable in portlet_plonemeeting
        self.assertFalse(searchAllItemsUID in self.portlet_pm_renderer.render())
        # but viewable in portlet_todo
        self.assertTrue(searchAllItemsUID in self.portlet_todo_renderer.render())
        self.assertTrue(item_url in self.portlet_todo_renderer.render())

    def test_pm_PortletTodoCaching(self):
        """Results updated when item modified or MeetingConfig changed."""
        cfg = self.meetingConfig
        cfg2 = self.meetingConfig2
        self._setup_portlets()
        # select 'searchallitems' in the MeetingConfig.toDoListSearches
        searchAllItemsUID = cfg.searches.searches_items.searchallitems.UID()
        cfg.setToDoListSearches([searchAllItemsUID])
        searchAllItemsUID = cfg2.searches.searches_items.searchallitems.UID()
        cfg2.setToDoListSearches([searchAllItemsUID])
        # enable rendering of portlet_todo, updated thru external view
        self.request.set('load_portlet_todo', True)

        # create items, one in each MeetingConfig
        self.changeUser('pmCreator1')
        item = self.create('MeetingItem')
        item_url = item.absolute_url()
        self.setMeetingConfig(cfg2.getId())
        item2 = self.create('MeetingItem')
        item2_url = item2.absolute_url()

        # MeetingConfig1
        self.assertTrue(item_url in self.portlet_todo_renderer.render())
        self.assertTrue(item.query_state() in self.portlet_todo_renderer.render())
        self.proposeItem(item)
        self.assertTrue(item_url in self.portlet_todo_renderer.render())
        self.assertTrue(item.query_state() in self.portlet_todo_renderer.render())

        # MeetingConfig1
        self.portlet_todo_renderer.cfg = cfg2
        self.assertTrue(item2_url in self.portlet_todo_renderer.render())
        self.assertTrue(item2.query_state() in self.portlet_todo_renderer.render())
        self.proposeItem(item2)
        self.assertTrue(item2_url in self.portlet_todo_renderer.render())
        self.assertTrue(item2.query_state() in self.portlet_todo_renderer.render())


def test_suite():
    from unittest import makeSuite
    from unittest import TestSuite
    suite = TestSuite()
    suite.addTest(makeSuite(testPortlets, prefix='test_pm_'))
    return suite



# -*- coding: utf-8 -*-
#
# File: MeetingCategory.py
#
# Copyright (c) 2017 by Imio.be
# Generator: ArchGenXML Version 2.7
#            http://plone.org/products/archgenxml
#
# GNU General Public License (GPL)
#

from AccessControl import ClassSecurityInfo
from imio.helpers.cache import invalidate_cachekey_volatile_for
from OFS.ObjectManager import BeforeDeleteException
from plone import api
from Products.Archetypes.atapi import AttributeStorage
from Products.Archetypes.atapi import BaseContent
from Products.Archetypes.atapi import BaseSchema
from Products.Archetypes.atapi import DisplayList
from Products.Archetypes.atapi import LinesField
from Products.Archetypes.atapi import MultiSelectionWidget
from Products.Archetypes.atapi import registerType
from Products.Archetypes.atapi import Schema
from Products.Archetypes.atapi import StringField
from Products.CMFCore.permissions import ModifyPortalContent
from Products.CMFDynamicViewFTI.browserdefault import BrowserDefaultMixin
from Products.PloneMeeting.config import PROJECTNAME
from Products.PloneMeeting.config import WriteRiskyConfig
from Products.PloneMeeting.utils import getCustomAdapter
from zope.i18n import translate
from zope.interface import implements

import interfaces


__author__ = """Gaetan DELANNAY <gaetan.delannay@geezteem.com>, Gauthier BASTIEN
<g.bastien@imio.be>, Stephan GEULETTE <s.geulette@imio.be>"""
__docformat__ = 'plaintext'


schema = Schema((

    StringField(
        name='categoryId',
        widget=StringField._properties['widget'](
            description="CategoryId",
            description_msgid="category_category_id_descr",
            label='Categoryid',
            label_msgid='PloneMeeting_label_categoryId',
            i18n_domain='PloneMeeting',
        ),
        write_permission="PloneMeeting: Write risky config",
        searchable=True,
    ),
    LinesField(
        name='usingGroups',
        widget=MultiSelectionWidget(
            description="UsingGroups",
            description_msgid="category_using_groups_descr",
            format="checkbox",
            label='Usinggroups',
            label_msgid='PloneMeeting_label_usingGroups',
            i18n_domain='PloneMeeting',
        ),
        enforceVocabulary=True,
        multiValued=1,
        vocabulary_factory='collective.contact.plonegroup.browser.settings.'
                           'SortedSelectedOrganizationsElephantVocabulary',
        write_permission="PloneMeeting: Write risky config",
    ),
    LinesField(
        name='categoryMappingsWhenCloningToOtherMC',
        widget=MultiSelectionWidget(
            description="CategoryMappingsWhenCloningToOtherMC",
            description_msgid="category_mapping_when_cloning_to_other_mc_descr",
            format="checkbox",
            label='Categorymappingswhencloningtoothermc',
            label_msgid='PloneMeeting_label_categoryMappingsWhenCloningToOtherMC',
            i18n_domain='PloneMeeting',
        ),
        multiValued=True,
        vocabulary='listCategoriesOfOtherMCs',
        enforceVocabulary=True,
        write_permission="PloneMeeting: Write risky config",
    ),
    LinesField(
        name='groupsInCharge',
        widget=MultiSelectionWidget(
            size=10,
            description="Groupsincharge",
            description_msgid="category_groups_in_charge_descr",
            format="checkbox",
            label='Groupsincharge',
            label_msgid='PloneMeeting_label_groupsInCharge',
            i18n_domain='PloneMeeting',
        ),
        multiValued=1,
        vocabulary_factory='collective.contact.plonegroup.browser.settings.'
                           'SortedSelectedOrganizationsElephantVocabulary',
    ),

),
)

MeetingCategory_schema = BaseSchema.copy() + \
    schema.copy()

# set write_permission for 'id' and 'title'
MeetingCategory_schema['id'].write_permission = "PloneMeeting: Write risky config"
MeetingCategory_schema['title'].write_permission = "PloneMeeting: Write risky config"
MeetingCategory_schema.changeSchemataForField('description', 'default')
MeetingCategory_schema.moveField('description', after='title')
MeetingCategory_schema['description'].storage = AttributeStorage()
MeetingCategory_schema['description'].write_permission = "PloneMeeting: Write risky config"
MeetingCategory_schema['description'].widget.description = " "
MeetingCategory_schema['description'].widget.description_msgid = "empty_description"
# hide metadata fields and even protect it by the WriteRiskyConfig permission
for field in MeetingCategory_schema.getSchemataFields('metadata'):
    field.widget.visible = {'edit': 'invisible', 'view': 'invisible'}
    field.write_permission = WriteRiskyConfig


class MeetingCategory(BaseContent, BrowserDefaultMixin):
    """
    """
    security = ClassSecurityInfo()
    implements(interfaces.IMeetingCategory)

    meta_type = 'MeetingCategory'
    _at_rename_after_creation = True

    schema = MeetingCategory_schema

    def isClassifier(self):
        '''Return True if current category is a classifier,
           False if it is a normal category.'''
        return self.getParentNode().getId() == 'classifiers'

    def getOrder(self, onlySelectable=True):
        '''At what position am I among all the active categories of my
           folder in the meeting config?  If p_onlySelectable is passed to
           MeetingConfig.getCategories, see doc string in MeetingConfig.'''
        try:
            # to avoid problems with categories that are disabled or
            # restricted to some groups, we pass onlySelectable=False
            tool = api.portal.get_tool('portal_plonemeeting')
            cfg = tool.getMeetingConfig(self)
            i = cfg.getCategories(
                classifiers=self.isClassifier(), onlySelectable=onlySelectable).index(self)
        except ValueError:
            i = None
        return i

    def _invalidateCachedVocabularies(self):
        """Clean cache for vocabularies using MeetingCategories."""
        invalidate_cachekey_volatile_for("Products.PloneMeeting.vocabularies.categoriesvocabulary")
        invalidate_cachekey_volatile_for("Products.PloneMeeting.vocabularies.groupsinchargevocabulary")

    security.declarePrivate('at_post_create_script')

    def at_post_create_script(self):
        self._invalidateCachedVocabularies()
        self.adapted().onEdit(isCreated=True)

    security.declarePrivate('at_post_edit_script')

    def at_post_edit_script(self):
        self._invalidateCachedVocabularies()
        self.adapted().onEdit(isCreated=False)

    security.declarePrivate('manage_beforeDelete')

    def manage_beforeDelete(self, item, container):
        '''Checks if the current MeetingCategory can be deleted:
          - it can not be linked to an existing meetingItem (normal item,
            recurring item or item template);
          - it can not be used in field 'categoryMappingsWhenCloningToOtherMC'
            of another MeetingCategory.'''
        return
        # If we are trying to remove the whole Plone Site, bypass this hook.
        # bypass also if we are in the creation process
        if not item.meta_type == "Plone Site" and not item._at_creation_flag:
            tool = api.portal.get_tool('portal_plonemeeting')
            cfg = tool.getMeetingConfig(self)
            catalog = api.portal.get_tool('portal_catalog')
            brains = catalog(
                portal_type=(
                    cfg.getItemTypeName(),
                    cfg.getItemTypeName(configType='MeetingItemRecurring'),
                    cfg.getItemTypeName(configType='MeetingItemTemplate')),
                getCategory=self.getId())
            if brains:
                # linked to an existing item, we can not delete it
                msg = translate(
                    "can_not_delete_meetingcategory_meetingitem",
                    domain="plone",
                    mapping={'url': brains[0].getURL()},
                    context=self.REQUEST)
                raise BeforeDeleteException(msg)
            # check field categoryMappingsWhenCloningToOtherMC of other MC categories
            cat_mapping_id = '{0}.{1}'.format(cfg.getId(), self.getId())
            for other_cfg in tool.objectValues('MeetingConfig'):
                if other_cfg == cfg:
                    continue
                for other_cat in other_cfg.getCategories(onlySelectable=False, caching=False):
                    if cat_mapping_id in other_cat.getCategoryMappingsWhenCloningToOtherMC():
                        msg = translate(
                            "can_not_delete_meetingcategory_other_category_mapping",
                            domain="plone",
                            mapping={'url': other_cat.absolute_url()},
                            context=self.REQUEST)
                        raise BeforeDeleteException(msg)
        BaseContent.manage_beforeDelete(self, item, container)

    security.declarePublic('getSelf')

    def getSelf(self):
        if self.getTagName() != 'MeetingCategory':
            return self.context
        return self

    security.declarePublic('adapted')

    def adapted(self):
        return getCustomAdapter(self)

    security.declareProtected(ModifyPortalContent, 'onEdit')

    def onEdit(self, isCreated):
        '''See doc in interfaces.py.'''
        pass

    security.declarePublic('isSelectable')

    def isSelectable(self, userId):
        '''See documentation in interfaces.py.'''
        cat = self.getSelf()
        tool = api.portal.get_tool('portal_plonemeeting')
        wfTool = api.portal.get_tool('portal_workflow')
        state = wfTool.getInfoFor(cat, 'review_state')
        isUsing = bool(state == 'active')
        usingGroups = cat.getUsingGroups()
        # If we have usingGroups make sure userId is creator for one of it
        if isUsing and usingGroups and not tool.isManager(cat, realManagers=True):
            cfg = tool.getMeetingConfig(cat)
            selectable_orgs = tool.get_selectable_orgs(cfg, user_id=userId)
            keys = [selectable_org.UID() for selectable_org in selectable_orgs]
            # Check intersection between self.usingGroups and orgs for which
            # the current user is creator
            isUsing = bool(set(usingGroups).intersection(keys))
        return isUsing

    security.declarePublic('listCategoriesOfOtherMCs')

    def listCategoriesOfOtherMCs(self):
        '''Vocabulary for 'categoryMappingsWhenCloningToOtherMC' field, it returns
           a list of available categories by available MC the items of the current MC
           can be sent to, like :
           - otherMC1 : category 1
           - otherMC1 : category 2
           - otherMC1 : category 3
           - otherMC1 : category 4
           - otherMC2 : category 1
           - otherMC2 : category 2'''
        res = []
        tool = api.portal.get_tool('portal_plonemeeting')
        cfg = tool.getMeetingConfig(self)
        # get every other MC the items of this MC can be sent to
        otherMCs = cfg.getMeetingConfigsToCloneTo()
        for otherMC in otherMCs:
            otherMCObj = getattr(tool, otherMC['meeting_config'])
            if otherMCObj.getUseGroupsAsCategories():
                continue
            otherMCId = otherMCObj.getId()
            otherMCTitle = otherMCObj.Title()
            for category in otherMCObj.getCategories(classifiers=self.isClassifier()):
                res.append(('%s.%s' % (otherMCId, category.getId()),
                            '%s -> %s' % (otherMCTitle, category.Title())))
        return DisplayList(tuple(res))

    security.declarePrivate('validate_categoryMappingsWhenCloningToOtherMC')

    def validate_categoryMappingsWhenCloningToOtherMC(self, values):
        '''This method does validate the 'categoryMappingsWhenCloningToOtherMC'.
           We can only select one single value (category) for a given MC.'''
        previousMCValue = 'DummyFalseMCId'
        for value in values:
            MCValue = value.split('.')[0]
            if MCValue.startswith(previousMCValue):
                return translate(u'error_can_not_select_several_cat_for_same_mc',
                                 domain="PloneMeeting",
                                 context=self.REQUEST)
            previousMCValue = MCValue


registerType(MeetingCategory, PROJECTNAME)



# -*- coding: utf-8 -*-
#
# File: testChangeItemOrderView.py
#
# GNU General Public License (GPL)
#

from AccessControl import Unauthorized
from Products.PloneMeeting.tests.PloneMeetingTestCase import PloneMeetingTestCase
from Products.PloneMeeting.utils import _storedItemNumber_to_itemNumber
from Products.statusmessages.interfaces import IStatusMessage
from zope.i18n import translate


class testChangeItemOrderView(PloneMeetingTestCase):
    '''Tests the functionnality that change items order on the meeting.'''

    def _setupOrderedItems(self):
        """ """
        meeting = self._createMeetingWithItems()
        # 7 items are created
        items = meeting.get_items(ordered=True)
        item1 = items[0]
        item2 = items[1]
        item3 = items[2]
        item4 = items[3]
        item5 = items[4]
        item6 = items[5]
        item7 = items[6]
        return meeting, item1, item2, item3, item4, item5, item6, item7

    def test_pm_ChangeItemOrderSetup(self):
        """As self._setupOrderedItems is used in several test methods, we added a test for it."""
        self.changeUser('pmManager')
        meeting, item1, item2, item3, item4, item5, item6, item7 = self._setupOrderedItems()
        self.assertEquals(item1.getItemNumber(), 100)
        self.assertEquals(item2.getItemNumber(), 200)
        self.assertEquals(item3.getItemNumber(), 300)
        self.assertEquals(item4.getItemNumber(), 400)
        self.assertEquals(item5.getItemNumber(), 500)
        self.assertEquals(item6.getItemNumber(), 600)
        self.assertEquals(item7.getItemNumber(), 700)
        self.assertEquals([item.getItemNumber() for item in meeting.get_items(ordered=True)],
                          [100, 200, 300, 400, 500, 600, 700])

    def test_pm_ChangeItemOrderIllegalMove(self):
        '''Illegal moves are :
           - moving < 1 or > last item;
           - valid number, like 4 or 4.1 or 4.13, no 4.123;
           - moving to existing subnumber, previous exist, so 2.3 needs 2.2;
           - master must exist for a subnumber, so moving to 12.1 needs 12;'''
        # create a meetingWithItems and play
        self.changeUser('pmManager')
        meeting, item1, item2, item3, item4, item5, item6, item7 = self._setupOrderedItems()
        view = item1.restrictedTraverse('@@change-item-order')
        # view returns False if not moved
        self.assertFalse(view('number', '0'))
        self.assertFalse(view('number', '0.1'))
        self.assertFalse(view('number', '9'))
        self.assertFalse(view('number', '8.1'))
        self.assertFalse(view('number', '2.2'))
        self.assertFalse(view('number', '2.23'))
        # move item to 2.1 so it exists then item2 to 1.3
        # (moving item 1 to 2.1 will make it have 1.1 number)
        view('number', '2.1')
        self.assertEqual(item1.getItemNumber(), 101)
        view = item5.restrictedTraverse('@@change-item-order')
        self.assertFalse(view('number', '1.3'))
        # moving last item to a too large number will not work
        self.assertEqual(item7.getItemNumber(), 600)
        view = item7.restrictedTraverse('@@change-item-order')
        self.assertFalse(view('number', '8'))
        # moving item 1.1 to 7 will work
        view = item7.restrictedTraverse('@@change-item-order')
        self.assertFalse(view('number', '8'))

    def test_pm_ChangeItemOrderMoveUpDown(self):
        '''Test the ChangeItemOrderView :
           - we can change an item one level up/down.'''
        # create a meetingWithItems and play
        self.changeUser('pmManager')
        meeting, item1, item2, item3, item4, item5, item6, item7 = self._setupOrderedItems()
        view = item1.restrictedTraverse('@@change-item-order')
        # move the item1 up, nothing changed...
        view('up')
        self.assertEquals(item1.getItemNumber(), 100)
        # move the item1 down, is position changed
        view('down')
        self.assertEquals(item1.getItemNumber(), 200)
        # and other items position are adapted
        self.assertEquals(item2.getItemNumber(), 100)
        self.assertEquals(item3.getItemNumber(), 300)
        self.assertEquals(item4.getItemNumber(), 400)
        self.assertEquals(item5.getItemNumber(), 500)
        self.assertEquals(item6.getItemNumber(), 600)
        self.assertEquals(item7.getItemNumber(), 700)
        # put back item1 to position 1
        view('up')
        # and other items position are adapted
        self.assertEquals(item1.getItemNumber(), 100)
        self.assertEquals(item2.getItemNumber(), 200)
        self.assertEquals(item3.getItemNumber(), 300)
        self.assertEquals(item4.getItemNumber(), 400)
        self.assertEquals(item5.getItemNumber(), 500)
        self.assertEquals(item6.getItemNumber(), 600)
        self.assertEquals(item7.getItemNumber(), 700)

    def test_pm_ChangeItemOrderMoveUpDownWithSubnumbers(self):
        '''Test the ChangeItemOrderView :
           - we may change subnumbers between them up/down but not between interger and subnumbers.'''
        self.changeUser('pmManager')
        meeting, item1, item2, item3, item4, item5, item6, item7 = self._setupOrderedItems()
        # move 300 to 201 and 400 to 202
        view = item3.restrictedTraverse('@@change-item-order')
        view('number', '2.1')
        view = item4.restrictedTraverse('@@change-item-order')
        view('number', '2.2')
        self.assertEquals([item.getItemNumber() for item in meeting.get_items(ordered=True)],
                          [100, 200, 201, 202, 300, 400, 500])
        # right, try to switch numbers 202 (item4) to 300
        self.assertEquals(item4.getItemNumber(), 202)
        view('down')
        self.assertEquals(item4.getItemNumber(), 202)
        # same while trying to change 201 with 200
        view = item3.restrictedTraverse('@@change-item-order')
        self.assertEquals(item3.getItemNumber(), 201)
        view('up')
        self.assertEquals(item3.getItemNumber(), 201)

        # may not switch neither 'interger' to 'subnumber'
        # trying to switch 300 and 202 will not change anything
        view = item5.restrictedTraverse('@@change-item-order')
        self.assertEquals(item5.getItemNumber(), 300)
        view('up')
        self.assertEquals(item5.getItemNumber(), 300)
        # no more trying to switch 200 and 201
        view = item2.restrictedTraverse('@@change-item-order')
        self.assertEquals(item2.getItemNumber(), 200)
        view('down')
        self.assertEquals(item2.getItemNumber(), 200)

        # but switching subnumbers between them works
        # switch 201 and 202
        view = item3.restrictedTraverse('@@change-item-order')
        self.assertEquals(item3.getItemNumber(), 201)
        self.assertEquals(item4.getItemNumber(), 202)
        view('down')
        self.assertEquals(item3.getItemNumber(), 202)
        self.assertEquals(item4.getItemNumber(), 201)

    def test_pm_ChangeItemOrderMoveIntegerToInteger(self):
        '''Test while moving up or down an integer item to another
           integer position (from 4 to 2 and 2 to 6 for example).'''
        self.changeUser('pmManager')
        meeting, item1, item2, item3, item4, item5, item6, item7 = self._setupOrderedItems()
        view = item4.restrictedTraverse('@@change-item-order')
        # move up
        view('number', '2')
        self.assertEquals(item1.getItemNumber(), 100)
        self.assertEquals(item2.getItemNumber(), 300)
        self.assertEquals(item3.getItemNumber(), 400)
        self.assertEquals(item4.getItemNumber(), 200)
        self.assertEquals(item5.getItemNumber(), 500)
        self.assertEquals(item6.getItemNumber(), 600)
        self.assertEquals(item7.getItemNumber(), 700)
        # ordered items are still right
        self.assertEquals([item.getItemNumber() for item in meeting.get_items(ordered=True)],
                          [100, 200, 300, 400, 500, 600, 700])
        # move down
        view('number', '6')
        self.assertEquals(item1.getItemNumber(), 100)
        self.assertEquals(item2.getItemNumber(), 200)
        self.assertEquals(item3.getItemNumber(), 300)
        self.assertEquals(item4.getItemNumber(), 600)
        self.assertEquals(item5.getItemNumber(), 400)
        self.assertEquals(item6.getItemNumber(), 500)
        self.assertEquals(item7.getItemNumber(), 700)
        # ordered items are still right
        self.assertEquals([item.getItemNumber() for item in meeting.get_items(ordered=True)],
                          [100, 200, 300, 400, 500, 600, 700])

    def test_pm_ChangeItemOrderMoveIntegerToSubnumber(self):
        '''Test while moving up or down an integer item to a subnumber
           position (from 4 to 2.1 and 3 to 6.1 for example).'''
        self.changeUser('pmManager')
        meeting, item1, item2, item3, item4, item5, item6, item7 = self._setupOrderedItems()
        view = item4.restrictedTraverse('@@change-item-order')
        # move up
        view('number', '2.1')
        self.assertEquals(item1.getItemNumber(), 100)
        self.assertEquals(item2.getItemNumber(), 200)
        self.assertEquals(item3.getItemNumber(), 300)
        self.assertEquals(item4.getItemNumber(), 201)
        self.assertEquals(item5.getItemNumber(), 400)
        self.assertEquals(item6.getItemNumber(), 500)
        self.assertEquals(item7.getItemNumber(), 600)
        self.assertEquals([item.getItemNumber() for item in meeting.get_items(ordered=True)],
                          [100, 200, 201, 300, 400, 500, 600])
        # move down
        # move 3 to 6.1
        view = item3.restrictedTraverse('@@change-item-order')
        view('number', '6.1')
        self.assertEquals(item1.getItemNumber(), 100)
        self.assertEquals(item2.getItemNumber(), 200)
        # as we have less numbers, 6.1 is sent to 5.1 actually
        self.assertEquals(item3.getItemNumber(), 501)
        self.assertEquals(item4.getItemNumber(), 201)
        self.assertEquals(item5.getItemNumber(), 300)
        self.assertEquals(item6.getItemNumber(), 400)
        self.assertEquals(item7.getItemNumber(), 500)
        self.assertEquals([item.getItemNumber() for item in meeting.get_items(ordered=True)],
                          [100, 200, 201, 300, 400, 500, 501])
        # move a "master number" down
        # move 2 to 4.1
        view = item2.restrictedTraverse('@@change-item-order')
        view('number', '4.1')
        self.assertEquals(item1.getItemNumber(), 100)
        # one masternumber less, so 4.1 became 3.1
        self.assertEquals(item2.getItemNumber(), 401)
        self.assertEquals(item3.getItemNumber(), 501)
        # does not have a master anymore, became master
        self.assertEquals(item4.getItemNumber(), 200)
        self.assertEquals(item5.getItemNumber(), 300)
        self.assertEquals(item6.getItemNumber(), 400)
        self.assertEquals(item7.getItemNumber(), 500)
        self.assertEquals([item.getItemNumber() for item in meeting.get_items(ordered=True)],
                          [100, 200, 300, 400, 401, 500, 501])
        # move a "master number" up
        # move 5 to 4.2
        view = item7.restrictedTraverse('@@change-item-order')
        view('number', '4.2')
        self.assertEquals(item1.getItemNumber(), 100)
        self.assertEquals(item2.getItemNumber(), 401)
        # does not have a master anymore, became master
        self.assertEquals(item3.getItemNumber(), 500)
        self.assertEquals(item4.getItemNumber(), 200)
        self.assertEquals(item5.getItemNumber(), 300)
        self.assertEquals(item6.getItemNumber(), 400)
        self.assertEquals(item7.getItemNumber(), 402)
        self.assertEquals([item.getItemNumber() for item in meeting.get_items(ordered=True)],
                          [100, 200, 300, 400, 401, 402, 500])

        # move up between 2 subnumbers
        # move 5 to 4.1
        view = item3.restrictedTraverse('@@change-item-order')
        view('number', '4.1')
        self.assertEquals([item.getItemNumber() for item in meeting.get_items(ordered=True)],
                          [100, 200, 300, 400, 401, 402, 403])

        # move down to an existing subnumber
        # move 2 to 4.1
        view = item4.restrictedTraverse('@@change-item-order')
        view('number', '4.1')
        self.assertEquals([item.getItemNumber() for item in meeting.get_items(ordered=True)],
                          [100, 200, 300, 301, 302, 303, 304])

    def test_pm_ChangeItemOrderMoveSubnumberToInteger(self):
        '''Test while moving up or down a subnumber to a subnumber (from 4.1 to 2 and 3.1 to 5 for example).'''
        self.changeUser('pmManager')
        meeting, item1, item2, item3, item4, item5, item6, item7 = self._setupOrderedItems()
        # prepare item numbers
        item3.setItemNumber(201)
        item4.setItemNumber(300)
        item5.setItemNumber(400)
        item6.setItemNumber(401)
        item7.setItemNumber(500)
        self.assertEquals(item1.getItemNumber(), 100)
        self.assertEquals(item2.getItemNumber(), 200)
        self.assertEquals(item3.getItemNumber(), 201)
        self.assertEquals(item4.getItemNumber(), 300)
        self.assertEquals(item5.getItemNumber(), 400)
        self.assertEquals(item6.getItemNumber(), 401)
        self.assertEquals(item7.getItemNumber(), 500)
        self.assertEquals([item.getItemNumber() for item in meeting.get_items(ordered=True)],
                          [100, 200, 201, 300, 400, 401, 500])
        # move up
        view = item6.restrictedTraverse('@@change-item-order')
        view('number', '2')
        self.assertEquals(item1.getItemNumber(), 100)
        self.assertEquals(item2.getItemNumber(), 300)
        self.assertEquals(item3.getItemNumber(), 301)
        self.assertEquals(item4.getItemNumber(), 400)
        self.assertEquals(item5.getItemNumber(), 500)
        self.assertEquals(item6.getItemNumber(), 200)
        self.assertEquals(item7.getItemNumber(), 600)
        # move down
        view = item3.restrictedTraverse('@@change-item-order')
        view('number', '5')
        self.assertEquals(item1.getItemNumber(), 100)
        self.assertEquals(item2.getItemNumber(), 300)
        self.assertEquals(item3.getItemNumber(), 500)
        self.assertEquals(item4.getItemNumber(), 400)
        self.assertEquals(item5.getItemNumber(), 600)
        self.assertEquals(item6.getItemNumber(), 200)
        self.assertEquals(item7.getItemNumber(), 700)

        # check with 2 groups of items with subnumbers
        item1.setItemNumber(100)
        item2.setItemNumber(200)
        item3.setItemNumber(201)
        item4.setItemNumber(300)
        item5.setItemNumber(400)
        item6.setItemNumber(500)
        item7.setItemNumber(501)
        self.assertEquals([item.getItemNumber() for item in meeting.get_items(ordered=True)],
                          [100, 200, 201, 300, 400, 500, 501])
        view = item3.restrictedTraverse('@@change-item-order')
        view('number', '3')
        self.assertEquals([item.getItemNumber() for item in meeting.get_items(ordered=True)],
                          [100, 200, 300, 400, 500, 600, 601])

        # move subnumber 300 to 602
        self.assertEqual(item3.getItemNumber(), 300)
        view = item3.restrictedTraverse('@@change-item-order')
        view('number', '6.2')
        self.assertEquals([item.getItemNumber() for item in meeting.get_items(ordered=True)],
                          [100, 200, 300, 400, 500, 501, 502])
        # trying to move 300 to 601 would do nothing
        self.assertEqual(item4.getItemNumber(), 300)
        view = item4.restrictedTraverse('@@change-item-order')
        view('number', '6.1')
        self.assertEquals([item.getItemNumber() for item in meeting.get_items(ordered=True)],
                          [100, 200, 300, 400, 500, 501, 502])
        # trying to move 502 to 601 would do nothing
        self.assertEqual(item3.getItemNumber(), 502)
        view = item3.restrictedTraverse('@@change-item-order')
        view('number', '6.1')
        self.assertEquals([item.getItemNumber() for item in meeting.get_items(ordered=True)],
                          [100, 200, 300, 400, 500, 501, 502])
        # trying to move 502 to 600 works
        self.assertEqual(item.getItemNumber(), 502)
        view = item.restrictedTraverse('@@change-item-order')
        view('number', '6')
        self.assertEquals([item.getItemNumber() for item in meeting.get_items(ordered=True)],
                          [100, 200, 300, 400, 500, 501, 600])
        # move 600 to 300
        self.assertEqual(item3.getItemNumber(), 600)
        view = item3.restrictedTraverse('@@change-item-order')
        view('number', '3')
        self.assertEquals([item.getItemNumber() for item in meeting.get_items(ordered=True)],
                          [100, 200, 300, 400, 500, 600, 601])
        # trying to move 300 to 603 would do nothing
        self.assertEqual(item3.getItemNumber(), 300)
        view = item3.restrictedTraverse('@@change-item-order')
        view('number', '6.3')
        self.assertEquals([item.getItemNumber() for item in meeting.get_items(ordered=True)],
                          [100, 200, 300, 400, 500, 600, 601])
        self.assertEqual(item3.getItemNumber(), 300)

        # now try to move last item
        # move 601 to 602, nothing changed
        self.assertEqual(item7.getItemNumber(), 601)
        view = item7.restrictedTraverse('@@change-item-order')
        view('number', '6.2')
        self.assertEquals([item.getItemNumber() for item in meeting.get_items(ordered=True)],
                          [100, 200, 300, 400, 500, 600, 601])
        # move 601 to 500 works
        self.assertEqual(item7.getItemNumber(), 601)
        view = item7.restrictedTraverse('@@change-item-order')
        view('number', '5')
        self.assertEqual(item7.getItemNumber(), 500)
        self.assertEquals([item.getItemNumber() for item in meeting.get_items(ordered=True)],
                          [100, 200, 300, 400, 500, 600, 700])
        # now we will move 601 to 700
        self.assertEqual(item6.getItemNumber(), 700)
        item6.setItemNumber(601)
        self.assertEquals([item.getItemNumber() for item in meeting.get_items(ordered=True)],
                          [100, 200, 300, 400, 500, 600, 601])
        view = item6.restrictedTraverse('@@change-item-order')
        view('number', '7')
        self.assertEqual(item6.getItemNumber(), 700)
        self.assertEquals([item.getItemNumber() for item in meeting.get_items(ordered=True)],
                          [100, 200, 300, 400, 500, 600, 700])

    def test_pm_ChangeItemOrderMoveSubnumberToSubnumber(self):
        '''Test while moving up or down a subnumber to a subnumber (from 4.1 to 2.1 and 3.1 to 5.1 for example).'''
        self.changeUser('pmManager')
        meeting, item1, item2, item3, item4, item5, item6, item7 = self._setupOrderedItems()
        # prepare item numbers
        item3.setItemNumber(201)
        item4.setItemNumber(300)
        item5.setItemNumber(400)
        item6.setItemNumber(401)
        item7.setItemNumber(500)
        self.assertEquals([item.getItemNumber() for item in meeting.get_items(ordered=True)],
                          [100, 200, 201, 300, 400, 401, 500])
        # move down
        view = item3.restrictedTraverse('@@change-item-order')
        view('number', '4.1')
        self.assertEquals(item1.getItemNumber(), 100)
        self.assertEquals(item2.getItemNumber(), 200)
        self.assertEquals(item3.getItemNumber(), 401)
        self.assertEquals(item4.getItemNumber(), 300)
        self.assertEquals(item5.getItemNumber(), 400)
        self.assertEquals(item6.getItemNumber(), 402)
        self.assertEquals(item7.getItemNumber(), 500)
        self.assertEquals([item.getItemNumber() for item in meeting.get_items(ordered=True)],
                          [100, 200, 300, 400, 401, 402, 500])
        # move up
        # prepare, change '500' to '201'
        item7.setItemNumber(201)
        meeting.notifyModified()
        self.assertEquals([item.getItemNumber() for item in meeting.get_items(ordered=True)],
                          [100, 200, 201, 300, 400, 401, 402])
        view = item3.restrictedTraverse('@@change-item-order')
        view('number', '2.1')
        self.assertEquals(item1.getItemNumber(), 100)
        self.assertEquals(item2.getItemNumber(), 200)
        self.assertEquals(item3.getItemNumber(), 201)
        self.assertEquals(item4.getItemNumber(), 300)
        self.assertEquals(item5.getItemNumber(), 400)
        self.assertEquals(item6.getItemNumber(), 401)
        self.assertEquals(item7.getItemNumber(), 202)
        self.assertEquals([item.getItemNumber() for item in meeting.get_items(ordered=True)],
                          [100, 200, 201, 202, 300, 400, 401])

        # move 300 to 401
        self.assertEquals(item4.getItemNumber(), 300)
        view = item4.restrictedTraverse('@@change-item-order')
        view('number', '4.1')
        self.assertEquals([item.getItemNumber() for item in meeting.get_items(ordered=True)],
                          [100, 200, 201, 202, 300, 301, 302])
        # move 201 to 302
        self.assertEquals(item3.getItemNumber(), 201)
        view = item3.restrictedTraverse('@@change-item-order')
        view('number', '3.2')
        self.assertEquals([item.getItemNumber() for item in meeting.get_items(ordered=True)],
                          [100, 200, 201, 300, 301, 302, 303])

    def test_pm_ChangeItemOrderMoveOutFromSubnumber(self):
        '''Test while moving an item back to an integer from a set of subnumbers, so we have :
           - 100, 200, 201, 202, 203, 204, 300 and we move 203 to 300 or 203 to 200.'''
        self.changeUser('pmManager')
        meeting, item1, item2, item3, item4, item5, item6, item7 = self._setupOrderedItems()
        # prepare item numbers
        item3.setItemNumber(201)
        item4.setItemNumber(202)
        item5.setItemNumber(203)
        item6.setItemNumber(204)
        item7.setItemNumber(300)
        self.assertEquals([item.getItemNumber() for item in meeting.get_items(ordered=True)],
                          [100, 200, 201, 202, 203, 204, 300])
        # move down 203 to 300
        view = item5.restrictedTraverse('@@change-item-order')
        view('number', '3')
        self.assertEquals([item.getItemNumber() for item in meeting.get_items(ordered=True)],
                          [100, 200, 201, 202, 203, 300, 400])
        self.assertEquals(item4.getItemNumber(), 202)
        self.assertEquals(item5.getItemNumber(), 300)
        self.assertEquals(item6.getItemNumber(), 203)
        self.assertEquals(item7.getItemNumber(), 400)

        # move up 202 to 100
        view = item4.restrictedTraverse('@@change-item-order')
        view('number', '1')
        self.assertEquals([item.getItemNumber() for item in meeting.get_items(ordered=True)],
                          [100, 200, 300, 301, 302, 400, 500])
        self.assertEquals(item4.getItemNumber(), 100)
        self.assertEquals(item1.getItemNumber(), 200)
        self.assertEquals(item2.getItemNumber(), 300)
        self.assertEquals(item6.getItemNumber(), 302)

    def test_pm_ChangeItemOrderMoveUpToFirstPositionWithSubnumbers(self):
        '''Test the ChangeItemOrderView :
           - we have 1, 1.1, 1.2, 2, 3, 4 and 5;
           - we move 3 to 1.
        '''
        # create a meetingWithItems and play
        self.changeUser('pmManager')
        meeting, item1, item2, item3, item4, item5, item6, item7 = self._setupOrderedItems()
        item1.setItemNumber(100)
        item2.setItemNumber(101)
        item3.setItemNumber(102)
        item4.setItemNumber(200)
        item5.setItemNumber(300)
        item6.setItemNumber(400)
        item7.setItemNumber(500)
        self.assertEquals([item.getItemNumber() for item in meeting.get_items(ordered=True)],
                          [100, 101, 102, 200, 300, 400, 500])
        view = item5.restrictedTraverse('@@change-item-order')
        view('number', '1')
        self.assertEquals([item.getItemNumber() for item in meeting.get_items(ordered=True)],
                          [100, 200, 201, 202, 300, 400, 500])

    def test_pm_ReorderItemsWithSubnumbers(self):
        '''When many subnumbers (1.1 to 1.6), move 1.2 to 1.4 and 1.5 to 1.2.'''
        self.changeUser('pmManager')
        meeting, item1, item2, item3, item4, item5, item6, item7 = self._setupOrderedItems()
        item1.setItemNumber(100)
        item2.setItemNumber(101)
        item3.setItemNumber(102)
        item4.setItemNumber(103)
        item5.setItemNumber(104)
        item6.setItemNumber(105)
        item7.setItemNumber(106)
        self.assertEquals([item.getItemNumber() for item in meeting.get_items(ordered=True)],
                          [100, 101, 102, 103, 104, 105, 106])
        # move down 1.2 to 1.4
        view = item3.restrictedTraverse('@@change-item-order')
        view('number', '1.4')
        # values still correct
        self.assertEquals([item.getItemNumber() for item in meeting.get_items(ordered=True)],
                          [100, 101, 102, 103, 104, 105, 106])
        self.assertEqual(item3.getItemNumber(), 104)
        self.assertEqual(item4.getItemNumber(), 102)
        self.assertEqual(item5.getItemNumber(), 103)

        # move up 1.4 to 1.2
        view = item3.restrictedTraverse('@@change-item-order')
        view('number', '1.2')
        # values still correct
        self.assertEquals([item.getItemNumber() for item in meeting.get_items(ordered=True)],
                          [100, 101, 102, 103, 104, 105, 106])
        self.assertEqual(item3.getItemNumber(), 102)
        self.assertEqual(item4.getItemNumber(), 103)
        self.assertEqual(item5.getItemNumber(), 104)

    def test_pm_ChangeItemOrderMoveAtGivenNumber(self):
        '''Test the ChangeItemOrderView :
           - we can change an item to a given p_moveNumber.'''
        # create a meetingWithItems and play
        self.changeUser('pmManager')
        meeting, item1, item2, item3, item4, item5, item6, item7 = self._setupOrderedItems()
        self.assertEquals(item1.getItemNumber(), 100)
        self.assertEquals(item2.getItemNumber(), 200)
        self.assertEquals(item3.getItemNumber(), 300)
        self.assertEquals(item4.getItemNumber(), 400)
        self.assertEquals(item5.getItemNumber(), 500)
        view = item2.restrictedTraverse('@@change-item-order')
        # move the item2 to position 3
        view('number', '3')
        self.assertEquals(item2.getItemNumber(), 300)
        # and other items position are adapted
        self.assertEquals(item1.getItemNumber(), 100)
        self.assertEquals(item3.getItemNumber(), 200)
        self.assertEquals(item4.getItemNumber(), 400)
        self.assertEquals(item5.getItemNumber(), 500)
        # put the item2 back to position 2
        view('up')
        self.assertEquals(item2.getItemNumber(), 200)
        # and other items position are adapted
        self.assertEquals(item1.getItemNumber(), 100)
        self.assertEquals(item3.getItemNumber(), 300)
        self.assertEquals(item4.getItemNumber(), 400)
        self.assertEquals(item5.getItemNumber(), 500)
        # no valid number does not cause crash
        view('number', '0')
        # nothing changed
        self.assertEquals(item2.getItemNumber(), 200)
        view('number', '-4')
        # nothing changed
        self.assertEquals(item2.getItemNumber(), 200)
        view('number', '99')
        # nothing changed
        self.assertEquals(item2.getItemNumber(), 200)
        view('number', None)
        # nothing changed
        self.assertEquals(item2.getItemNumber(), 200)
        # move 2 positions to far
        self.assertEqual(len(meeting.get_items()), 7)
        view('number', '9')
        # nothing changed
        self.assertEquals(item2.getItemNumber(), 200)
        # move one of the last items upper
        view = item4.restrictedTraverse('@@change-item-order')
        view('number', '1')
        self.assertEquals(item4.getItemNumber(), 100)
        # and other items position are adapted
        self.assertEquals(item1.getItemNumber(), 200)
        self.assertEquals(item2.getItemNumber(), 300)
        self.assertEquals(item3.getItemNumber(), 400)
        self.assertEquals(item5.getItemNumber(), 500)
        # change the item to the same place, a message is displayed
        messages = IStatusMessage(self.request)
        view('number', _storedItemNumber_to_itemNumber(item4.getItemNumber()))
        self.assertEquals(messages.show()[-1].message, translate('item_did_not_move', 'PloneMeeting'))

    def test_pm_MoveLateItemDoNotChangeNormalItems(self):
        """
          Normal items are moved between them and late items also.
        """
        # create a meetingWithItems and play
        self.changeUser('pmManager')
        meeting = self._createMeetingWithItems()
        for item in meeting.get_items():
            item.setDecision('<p>Dummy decision</p>')
        # freeze the meeting to be able to add late items
        self.freezeMeeting(meeting)
        # create 4 items that will be late
        late1 = self.create('MeetingItem')
        late1.setPreferredMeeting(meeting.UID())
        late1.reindexObject()
        late2 = self.create('MeetingItem')
        late2.setPreferredMeeting(meeting.UID())
        late2.setTitle('i2')
        late2.reindexObject()
        late3 = self.create('MeetingItem')
        late3.setPreferredMeeting(meeting.UID())
        late3.setTitle('i3')
        late3.reindexObject()
        late4 = self.create('MeetingItem')
        late4.setPreferredMeeting(meeting.UID())
        late4.setTitle('i3')
        late4.reindexObject()
        # present the items
        for item in (late1, late2, late3, late4):
            self.presentItem(item)
        item1 = meeting.get_items(ordered=True)[0]
        item2 = meeting.get_items(ordered=True)[1]
        item3 = meeting.get_items(ordered=True)[2]
        item4 = meeting.get_items(ordered=True)[3]
        # normal and late items manage their own order
        self.assertEquals(item1.getItemNumber(), 100)
        self.assertEquals(item2.getItemNumber(), 200)
        self.assertEquals(item3.getItemNumber(), 300)
        self.assertEquals(item4.getItemNumber(), 400)
        self.assertEquals(late1.getItemNumber(), 500)
        self.assertEquals(late2.getItemNumber(), 600)
        self.assertEquals(late3.getItemNumber(), 700)
        self.assertEquals(late4.getItemNumber(), 800)
        # move a late item and check that normal items are not changed
        view = late2.restrictedTraverse('@@change-item-order')
        view('up')
        # late2 position changed but not normal items
        self.assertEquals(item1.getItemNumber(), 100)
        self.assertEquals(item2.getItemNumber(), 200)
        self.assertEquals(item3.getItemNumber(), 300)
        self.assertEquals(item4.getItemNumber(), 400)
        self.assertEquals(late1.getItemNumber(), 600)
        self.assertEquals(late2.getItemNumber(), 500)
        self.assertEquals(late3.getItemNumber(), 700)
        self.assertEquals(late4.getItemNumber(), 800)

    def test_pm_MayChangeItemOrder(self):
        """
          The item order can be changed until Meeting.may_change_items_order is False
        """
        # create a meetingWithItems and play
        self.changeUser('pmManager')
        meeting = self._createMeetingWithItems()
        item = meeting.get_items(ordered=True)[0]
        view = item.restrictedTraverse('@@change-item-order')
        self.assertTrue(meeting.wfConditions().may_change_items_order())
        view('down')
        self.freezeMeeting(meeting)
        self.assertTrue(meeting.wfConditions().may_change_items_order())
        view('down')
        # add decision to items so meeting can be decided
        for item in meeting.get_items():
            item.setDecision('<p>Dummy decision</p>')
            item.reindexObject(idxs=['getDecision', ])
        # items order is changeable until the meeting is in a closed state
        for tr in self._getTransitionsToCloseAMeeting():
            if tr in self.transitions(meeting):
                self.do(meeting, tr)
                # order still changeable
                if not meeting.query_state() in meeting.MEETINGCLOSEDSTATES:
                    self.assertTrue(meeting.wfConditions().may_change_items_order())
                else:
                    # if the meeting is in a closed state, order is no more changeable
                    self.assertFalse(meeting.wfConditions().may_change_items_order())
                    # if may_change_items_order is False, trying to change
                    # order will raise an Unauthorized
                    self.assertRaises(Unauthorized, view, 'up')
                    self.assertRaises(Unauthorized, view, 'down')

    def test_pm_ChangeItemOrderMoveToLastPosition(self):
        '''Move item to last position :
           - from 3 to last position or last position + 1;
           - from 3.1 to last position or last position + 1;
           - from 3.1 to 4.2;'''
        # create a meetingWithItems and play
        self.changeUser('pmManager')
        meeting, item1, item2, item3, item4, item5, item6, item7 = self._setupOrderedItems()
        view = item1.restrictedTraverse('@@change-item-order')
        # integer to integer
        view('number', '7')
        self.assertEqual(item1.getItemNumber(), 700)
        view('number', '1')
        view('number', '8')
        self.assertEqual(item1.getItemNumber(), 700)

        # integer to subnumber
        # move 1 to 7.1
        view('number', '1')
        view('number', '7.1')
        self.assertEqual(item1.getItemNumber(), 601)

        # subnumber to subnumber
        view = item4.restrictedTraverse('@@change-item-order')
        view('number', '2.1')
        view('number', '5.2')
        self.assertEqual(item4.getItemNumber(), 502)

        # integer to integer with subnumber as last item
        view = item5.restrictedTraverse('@@change-item-order')
        view('number', '6')
        self.assertEqual(item5.getItemNumber(), 500)
        view('number', '2.1')
        # will not move if too far
        self.assertFalse(view('number', '6'))


def test_suite():
    from unittest import makeSuite
    from unittest import TestSuite
    suite = TestSuite()
    suite.addTest(makeSuite(testChangeItemOrderView, prefix='test_pm_'))
    return suite



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
# GNU General Public License (GPL)
# ------------------------------------------------------------------------------
'''This module defines functions that allow to migrate to a given version of
   PloneMeeting for production sites that run older versions of PloneMeeting.
   You must run every migration function in the right chronological order.'''

from collections import OrderedDict
from collective.behavior.talcondition.behavior import ITALCondition
from collective.documentgenerator.search_replace.pod_template import SearchAndReplacePODTemplates
from collective.iconifiedcategory.utils import _categorized_elements
from DateTime import DateTime
from eea.facetednavigation.interfaces import ICriteria
from imio.helpers.cache import cleanRamCacheFor
from imio.helpers.catalog import addOrUpdateColumns
from imio.helpers.catalog import addOrUpdateIndexes
from imio.helpers.content import object_values
from imio.helpers.content import uuidToObject
from imio.migrator.migrator import Migrator as BaseMigrator
from imio.pyutils.utils import replace_in_list
from natsort import humansorted
from operator import attrgetter
from plone import api
from Products.Archetypes.event import ObjectEditedEvent
from Products.CMFPlone.utils import base_hasattr
from Products.PloneMeeting.content.meeting import IMeeting
from Products.PloneMeeting.MeetingConfig import ITEM_WF_STATE_ATTRS
from Products.PloneMeeting.MeetingConfig import ITEM_WF_TRANSITION_ATTRS
from Products.PloneMeeting.MeetingConfig import MEETING_WF_STATE_ATTRS
from Products.PloneMeeting.MeetingConfig import MEETING_WF_TRANSITION_ATTRS
from Products.PloneMeeting.setuphandlers import columnInfos
from Products.PloneMeeting.setuphandlers import indexInfos
from Products.PloneMeeting.utils import forceHTMLContentTypeForEmptyRichFields
from Products.PloneMeeting.utils import reindex_object
from Products.ZCatalog.ProgressHandler import ZLogHandler
from zope.event import notify
from zope.i18n import translate

import itertools
import logging
import os


logger = logging.getLogger('PloneMeeting')


class Migrator(BaseMigrator):
    '''Abstract class for creating a migrator.'''

    already_migrated = False

    def __init__(self, context):
        BaseMigrator.__init__(self, context, disable_linkintegrity_checks=True)
        self.tool = api.portal.get_tool('portal_plonemeeting')
        # disable email notifications for every MeetingConfigs and save
        # current state to set it back after migration in self.finish
        self.cfgsMailMode = {}
        for cfg in self.tool.objectValues('MeetingConfig'):
            self.cfgsMailMode[cfg.getId()] = cfg.getMailMode()
            cfg.setMailMode('deactivated')
        # disable advices invalidation for every MeetingConfigs and save
        # current state to set it back after migration in self.finish
        self.cfgsAdvicesInvalidation = {}
        for cfg in self.tool.objectValues('MeetingConfig'):
            self.cfgsAdvicesInvalidation[cfg.getId()] = cfg.getEnableAdviceInvalidation()
            cfg.setEnableAdviceInvalidation(False)
        self.profile_name = u'profile-Products.PloneMeeting:default'

    def reorderSkinsLayers(self):
        """Reapply skins of Products.PloneMeeting + self.profile_name."""
        # re-apply the PloneMeeting skins and the self.profile_name skin if different
        self.runProfileSteps('Products.PloneMeeting',
                             steps=['skins'],
                             profile='default')
        if self.profile_name != u'profile-Products.PloneMeeting:default':
            product_name = self.profile_name.split(':')[0][8:]
            self.runProfileSteps(product_name,
                                 steps=['skins'],
                                 profile='default')

    def upgradeDependencies(self):
        """Upgrade every dependencies."""
        profile_names = self.ps.getDependenciesForProfile(u'profile-Products.PloneMeeting:default')
        for profile_name in profile_names:
            self.upgradeProfile(profile_name)
        if self.profile_name != 'profile-Products.PloneMeeting:default':
            profile_names = self.ps.getDependenciesForProfile(self.profile_name)
            for profile_name in profile_names:
                self.upgradeProfile(profile_name)

    def reinstall(self, profiles, ignore_dependencies=False, dependency_strategy=None):
        """Override to be able to call _after_reinstall at the end."""
        self._before_reinstall()
        BaseMigrator.reinstall(self, profiles, ignore_dependencies, dependency_strategy)
        self._after_reinstall()

    def _before_reinstall(self):
        """Before self.reinstall hook that let's a subplugin knows that the profile
           will be executed and may launch some migration steps before PM ones."""
        # save CKeditor custom styles
        cke_props = self.portal.portal_properties.ckeditor_properties
        self.menuStyles = cke_props.menuStyles

    def _after_reinstall(self):
        """After self.reinstall hook that let's a subplugin knows that the profile
           has been executed and may launch some migration steps before PM ones."""
        # set back CKeditor custom styles
        cke_props = self.portal.portal_properties.ckeditor_properties
        cke_props.menuStyles = self.menuStyles

    def getWorkflows(self, meta_types=['Meeting',
                                       'MeetingItem',
                                       'MeetingItemTemplate',
                                       'MeetingItemRecurring']):
        """Returns every workflows used for every portal_types based on given p_meta_type."""
        portal_types = []
        for cfg in self.tool.objectValues('MeetingConfig'):
            for meta_type in meta_types:
                if meta_type == 'Meeting':
                    portal_types.append(cfg.getMeetingTypeName())
                elif meta_type == 'MeetingItem':
                    portal_types.append(cfg.getItemTypeName())
                else:
                    # MeetingItemXXX type
                    portal_types.append(cfg.getItemTypeName(configType=meta_type))
        wf_ids = [self.wfTool.getWorkflowsFor(portal_type)[0].getId()
                  for portal_type in portal_types]
        return wf_ids

    def updateWFStatesAndTransitions(self,
                                     related_to='MeetingItem',
                                     query={},
                                     review_state_mappings={},
                                     transition_mappings={},
                                     update_local_roles=False):
        """Update for given p_brains the workflow_history keys 'review_state' and 'action'
           depending on given p_review_state_mappings and p_action_mappings.
           Update also various parameters of the MeetingConfig
           that are using states and transitions."""
        logger.info(
            'Updating workflow states/transitions changes for elements of type "%s"...'
            % query or related_to)

        def _replace_values_in_list(values, review_state_mappings):
            """ """
            for original, replacement in review_state_mappings.items():
                values = replace_in_list(values, original, replacement)
                # try also to replace a value like 'Meeting.frozen'
                original = '%s.%s' % (related_to, original)
                replacement = '%s.%s' % (related_to, replacement)
                values = replace_in_list(values, original, replacement)
            return values

        def _update_attrs(attrs, mappings):
            for attr in attrs:
                # state_attr may be a datagridfield field_name/column_name
                # like powerObservers/item_states
                if "/" in attr:
                    field_name, col_name = attr.split('/')
                    values = cfg.getField(field_name).get(cfg)
                    for row in values:
                        if hasattr(row[col_name], '__iter__'):
                            col_values = _replace_values_in_list(
                                row[col_name], mappings)
                        else:
                            col_values = mappings.get(row[col_name], row[col_name])
                        row[col_name] = col_values
                else:
                    field_name = attr
                    values = cfg.getField(field_name).get(cfg)
                    values = _replace_values_in_list(values, mappings)
                setattr(cfg, field_name, tuple(values))

        # MeetingConfigs
        state_attrs = ITEM_WF_STATE_ATTRS if related_to == 'MeetingItem' else MEETING_WF_STATE_ATTRS
        tr_attrs = ITEM_WF_TRANSITION_ATTRS if related_to == 'MeetingItem' else MEETING_WF_TRANSITION_ATTRS
        for cfg in self.tool.objectValues('MeetingConfig'):
            # state_attrs
            _update_attrs(state_attrs, review_state_mappings)

            # transition_attrs
            _update_attrs(tr_attrs, transition_mappings)

        # workflow_history
        # manage query if not given
        if not query:
            if related_to == 'MeetingItem':
                query = {'meta_type': 'MeetingItem'}
            else:
                query = {'object_provides': IMeeting.__identifier__}
        brains = self.portal.portal_catalog(**query)
        pghandler = ZLogHandler(steps=1000)
        pghandler.init('Updating workflow_history', len(brains))
        i = 0
        objsToUpdate = []
        for brain in brains:
            i += 1
            pghandler.report(i)
            itemOrMeeting = brain.getObject()
            for wf_name, events in itemOrMeeting.workflow_history.items():
                for event in events:
                    if event['review_state'] in review_state_mappings:
                        event['review_state'] = review_state_mappings[event['review_state']]
                        itemOrMeeting.workflow_history._p_changed = True
                        objsToUpdate.append(itemOrMeeting)
                    if event['action'] in transition_mappings:
                        event['action'] = transition_mappings[event['action']]
                        itemOrMeeting.workflow_history._p_changed = True
                        # not necessary if just an action changed?
                        # objsToUpdate.append(itemOrMeeting)
        # update fixed objects
        if update_local_roles:
            for obj in objsToUpdate:
                obj.update_local_roles()
                # use reindex_object and pass some no_idxs because
                # calling reindexObject will update modified
                reindex_object(obj, no_idxs=['SearchableText', 'Title', 'Description'])

    def addCatalogIndexesAndColumns(self, indexes=True, columns=True, update_metadata=True):
        """ """
        if indexes:
            addOrUpdateIndexes(self.portal, indexInfos)
        if columns:
            addOrUpdateColumns(self.portal, columnInfos, update_metadata=update_metadata)

    def updateTALConditions(self, old_word, new_word):
        """Update every elements having a tal_condition, replace given old_word by new_word."""
        logger.info('Updating TAL conditions, replacing "{0}" by "{1}"'.format(old_word, new_word))
        # ITALConditionable : DashboardCollection, PODTemplates
        for brain in api.content.find(
                object_provides='collective.behavior.talcondition.interfaces.ITALConditionable'):
            obj = brain.getObject()
            adapted = ITALCondition(obj)
            tal_condition = adapted.get_tal_condition()
            if tal_condition and old_word in tal_condition:
                tal_condition = tal_condition.replace(old_word, new_word)
                adapted.set_tal_condition(tal_condition)
                logger.info('Word "{0}" was replaced by "{1}" for element "{2}"'.format(
                    old_word, new_word, repr(obj)))
            # mailing_lists for POD templates
            if obj.__class__.__name__ == "ConfigurablePODTemplate":
                mailing_lists = obj.mailing_lists
                if mailing_lists and old_word in mailing_lists:
                    mailing_lists = mailing_lists.replace(old_word, new_word)
                    obj.mailing_lists = mailing_lists
                    logger.info('Word "{0}" was replaced by "{1}" for element "{2}"'.format(
                        old_word, new_word, repr(obj)))

        # MeetingConfig
        for cfg in object_values(self.tool, 'MeetingConfig'):
            # datagrid fields
            # column names holding TAL expressions
            datagrid_tal_fields = ['tal_expression',
                                   'gives_auto_advice_on',
                                   'available_on',
                                   'item_access_on',
                                   'meeting_access_on']
            # datagrid fields holding TAL expressions
            datagrid_fields = ["onTransitionFieldTransforms",
                               "onMeetingTransitionItemActionToExecute",
                               "customAdvisers",
                               "powerObservers"]
            for datagrid_fieldname in datagrid_fields:
                adapted_value = getattr(cfg, datagrid_fieldname)
                for row in adapted_value:
                    for datagrid_tal_field in datagrid_tal_fields:
                        if datagrid_tal_field in row:
                            row[datagrid_tal_field] = \
                                row[datagrid_tal_field].replace(old_word, new_word)
                setattr(cfg, datagrid_fieldname, adapted_value)
            # other fields
            for field_name in ["itemReferenceFormat", "voteCondition"]:
                field = cfg.getField(field_name)
                value = field.get(cfg)
                value = value.replace(old_word, new_word)
                field.set(cfg, value)
        # organizations
        for brain in self.catalog(portal_type="organization"):
            org = brain.getObject()
            as_copy_group_on = getattr(org, "as_copy_group_on", None)
            if as_copy_group_on is not None:
                as_copy_group_on = as_copy_group_on.replace(old_word, new_word)
                org.as_copy_group_on = as_copy_group_on
        logger.info('Done.')

    def updatePODTemplatesCode(self,
                               replacements={},
                               meeting_replacements={},
                               item_replacements={}):
        """Apply given p_replacements to every POD templates.
           p_meeting_replacements are for POD templates used on Meetings and
           p_item_replacements are for POD templates used on MeetingItems.
           This let know what "self" is, a meeting or an item.
           WARNING, be defensive with replacements :
           - try to start with "self", "." or "=" and end with "(" or "()",
             never use single word, a single word may also be a
             POD template context variable ("listTypes", "review_state", ...)."""
        logger.info('Fixing POD templates instructions....')
        # first make sure collective.documentgenerator is upgraded
        # before search&replace in POD templates
        self.upgradeProfile("collective.documentgenerator:default")
        results = []
        for cfg in self.tool.objectValues('MeetingConfig'):
            pod_templates = object_values(
                cfg.podtemplates,
                ['ConfigurablePODTemplate', 'DashboardPODTemplate'])
            # only keep POD templates having an odt_file
            pod_templates = [pt for pt in pod_templates if pt.odt_file]
            meeting_type_name = cfg.getMeetingTypeName()
            item_type_name = cfg.getItemTypeName()
            for pod_template in pod_templates:
                logger.info('Checking POD template at %s...' % repr(pod_template))
                with SearchAndReplacePODTemplates([pod_template]) as search_replace:
                    # apply first portal_type specific replacements (Meeting or Item)
                    if pod_template.pod_portal_types:
                        if meeting_type_name in pod_template.pod_portal_types:
                            for k, v in meeting_replacements.items():
                                res = search_replace.replace(k, v, is_regex=False)
                                if res:
                                    results.append(res)
                        if item_type_name in pod_template.pod_portal_types:
                            for k, v in item_replacements.items():
                                res = search_replace.replace(k, v, is_regex=False)
                                if res:
                                    results.append(res)
                    # replacements compatible with any portal_types
                    for k, v in replacements.items():
                        res = search_replace.replace(k, v, is_regex=False)
                        if res:
                            results.append(res)
        # format results and dump it in the Zope log
        # as clean as possible so it can be used to know what changed
        data = {}
        for result in results:
            pt_uid, infos = result.items()[0]
            pt = uuidToObject(pt_uid, unrestricted=True)
            pt_path_and_title = "{0} - {1}".format(
                '/'.join(pt.getPhysicalPath()), pt.Title())
            if pt_path_and_title not in data:
                data[pt_path_and_title] = []
                self.warnings.append('Replacements were done in POD template at %s'
                                     % pt_path_and_title)
            for info in infos:
                # collective.documentgenerator < 3.30 from which we use appy.pod S&R
                # XXX to be removed when using collective.documentgenerator >= 3.30
                if hasattr(info, 'pod_expr'):
                    data[pt_path_and_title].append("---- " + info.pod_expr)
                    data[pt_path_and_title].append("++++ " + info.new_pod_expr)
                else:
                    line = repr(info).replace('  These changes were done:', '>>>'). \
                        replace('\n\n', '\n').rstrip('\n')
                    data[pt_path_and_title].append(line)
        logger.info("REPLACEMENTS IN POD TEMPLATES")
        if not data:
            logger.info("=============================")
            logger.info("No replacement was done.")
        else:
            # order data by pt_path
            ordered_data = OrderedDict(sorted(data.items()))
            output = ["============================="]
            for pt_path_and_title, infos in ordered_data.items():
                output.append("POD template " + pt_path_and_title)
                output.append('-' * len("POD template " + pt_path_and_title))
                for info in infos:
                    output.append(info)
            # make sure we do not mix unicode and utf-8
            fixed_output = []
            for line in output:
                if isinstance(line, unicode):
                    line = line.encode('utf8')
                fixed_output.append(line)
            logger.info('\n'.join(fixed_output))
        logger.info('Done.')

    def updateHolidays(self):
        '''Update holidays using default holidays.'''
        logger.info('Updating holidays...')
        from Products.PloneMeeting.profiles import PloneMeetingConfiguration
        defaultPMConfig = PloneMeetingConfiguration('', '', '')
        defaultHolidays = [holiday['date'] for holiday in defaultPMConfig.holidays]
        currentHolidays = [holiday['date'] for holiday in self.tool.getHolidays()]
        storedHolidays = list(self.tool.getHolidays())
        highestStoredHoliday = DateTime(storedHolidays[-1]['date'])
        for defaultHoliday in defaultHolidays:
            # update if not there and if higher that highest stored holiday
            if defaultHoliday not in currentHolidays and \
               DateTime(defaultHoliday) > highestStoredHoliday:
                storedHolidays.append({'date': defaultHoliday})
                logger.info('Adding {0} to holidays'.format(defaultHoliday))
        self.tool.setHolidays(storedHolidays)
        logger.info('Done.')

    def addNewSearches(self):
        """Add new searches by createSearches."""
        logger.info('Adding new searches...')
        for cfg in self.tool.objectValues('MeetingConfig'):
            cfg._createSubFolders()
            cfg.createSearches(cfg._searchesInfo())
            cfg.updateCollectionColumns()
        logger.info('Done.')

    def updateCollectionColumns(self):
        """Update collections columns."""
        logger.info("Updating collections columns for every MeetingConfigs...")
        for cfg in self.tool.objectValues('MeetingConfig'):
            cfg.updateCollectionColumns()
        logger.info('Done.')

    def cleanMeetingConfigs(self, field_names=[], renamed={}):
        """Remove given p_field_names from every MeetingConfigs.
           If a field has same type but was just renamed,
           it may be given in p_renamed dict mapping."""
        logger.info('Cleaning MeetingConfigs...')
        for cfg in self.tool.objectValues('MeetingConfig'):
            for field_name in field_names:
                if base_hasattr(cfg, field_name):
                    if field_name in renamed:
                        setattr(cfg, renamed[field_name], getattr(cfg, field_name))
                    delattr(cfg, field_name)
        logger.info('Done.')

    def cleanTool(self, field_names=[]):
        """Remove given p_field_names from ToolPloneMeeting."""
        logger.info('Cleaning ToolPloneMeeting...')
        for field_name in field_names:
            if base_hasattr(self.tool, field_name):
                delattr(self.tool, field_name)
        logger.info('Done.')

    def initNewHTMLFields(self, query={'meta_type': 'MeetingItem'}, field_names=[]):
        '''Make sure the content_type is correctly set to 'text/html' for new xhtml fields.'''
        logger.info('Initializing new HTML fields (query=%s, field_names=%s)...'
                    % (query, field_names))
        brains = self.portal.portal_catalog(**query)
        pghandler = ZLogHandler(steps=1000)
        pghandler.init('Initializing new HTML fields', len(brains))
        i = 0
        for brain in brains:
            i += 1
            pghandler.report(i)
            itemOrMeeting = brain.getObject()
            if field_names:
                for field_name in field_names:
                    forceHTMLContentTypeForEmptyRichFields(itemOrMeeting, field_name=field_name)
            else:
                forceHTMLContentTypeForEmptyRichFields(itemOrMeeting)
        pghandler.finish()
        logger.info('Done.')

    def updateColumns(self,
                      field_names=('itemColumns',
                                   'availableItemsListVisibleColumns',
                                   'itemsListVisibleColumns',
                                   'meetingColumns'),
                      to_remove=[],
                      to_replace={},
                      to_add=[],
                      cfg_ids=[]):
        '''When a column is no more available.'''
        logger.info('Cleaning MeetingConfig columns related fields, '
                    'removing columns "%s", replacing "%s" and adding "%s"...'
                    % (', '.join(to_remove),
                       ', '.join(to_replace.keys()),
                       ', '.join(to_add)))
        for cfg in self.tool.objectValues('MeetingConfig'):
            if cfg_ids and cfg.getId() not in cfg_ids:
                continue
            for field_name in field_names:
                field = cfg.getField(field_name)
                keys = field.get(cfg)
                adapted_keys = [k for k in keys if k not in to_remove]
                for orignal_value, new_value in to_replace.items():
                    adapted_keys = replace_in_list(adapted_keys, orignal_value, new_value)
                for col_to_add in to_add:
                    if col_to_add not in adapted_keys:
                        adapted_keys.append(col_to_add)
                field.set(cfg, adapted_keys)
            cfg.updateCollectionColumns()
        logger.info('Done.')

    def update_faceted_filters(self,
                               to_remove=[],
                               to_add=[],
                               field_names=['dashboardItemsListingsFilters',
                                            'dashboardMeetingAvailableItemsFilters',
                                            'dashboardMeetingLinkedItemsFilters'],
                               cfg_ids=[]):
        '''When a faceted filter is no more available or has been renamed.'''
        logger.info('Updating MeetingConfig faceted filter related fields, '
                    'removing filters "%s" and adding filters "%s"...'
                    % (', '.join(to_remove),
                       ', '.join(to_add)))
        for cfg in self.tool.objectValues('MeetingConfig'):
            if cfg_ids and cfg.getId() not in cfg_ids:
                continue
            for field_name in field_names:
                field = cfg.getField(field_name)
                keys = field.get(cfg)
                adapted_keys = [k for k in keys if k not in to_remove]
                for filter_to_add in to_add:
                    if filter_to_add not in adapted_keys:
                        adapted_keys.append(filter_to_add)
                field.set(cfg, adapted_keys)
        logger.info('Done.')

    def update_used_attrs(self, to_remove=[], to_add=[], to_replace={}, cfg_ids=[], related_to="MeetingItem"):
        '''When an item attribute is no more available or has been renamed.'''
        logger.info('Updating MeetingConfig "%s" used attributes, '
                    'removing "%s", replacing "%s" and adding "%s"...'
                    % (related_to,
                       ', '.join(to_remove),
                       ', '.join(to_replace.keys()),
                       ', '.join(to_add)))
        for cfg in self.tool.objectValues('MeetingConfig'):
            if cfg_ids and cfg.getId() not in cfg_ids:
                continue
            if related_to == "MeetingItem":
                used_attrs = list(cfg.getUsedItemAttributes())
            else:
                used_attrs = list(cfg.getUsedMeetingAttributes())
            adapted_used_attrs = [k for k in used_attrs if k not in to_remove]
            for attr_to_add in to_add:
                if attr_to_add not in adapted_used_attrs:
                    adapted_used_attrs.append(attr_to_add)
            for orignal_value, new_value in to_replace.items():
                adapted_used_attrs = replace_in_list(adapted_used_attrs, orignal_value, new_value)

            if related_to == "MeetingItem":
                cfg.setUsedItemAttributes(adapted_used_attrs)
            else:
                cfg.setUsedMeetingAttributes(adapted_used_attrs)
        logger.info('Done.')

    def removeUnusedWorkflows(self):
        '''Check used workflows and remove workflows containing '__' that are not used.'''
        logger.info('Cleaning unused workflows...')
        used_workflows = [wf_ids for portal_type_id, wf_ids in
                          self.wfTool._chains_by_type.items() if wf_ids]
        pm_workflows = tuple(set(sorted([wf_id for wf_id in
                                         itertools.chain.from_iterable(used_workflows)
                                         if '__' in wf_id])))
        to_delete = [wf_id for wf_id in self.wfTool.getWorkflowIds()
                     if '__' in wf_id and wf_id not in pm_workflows]
        if to_delete:
            logger.warning('Removing following workflows: "%s"' % ', '.join(to_delete))
            self.wfTool.manage_delObjects(to_delete)
        logger.info('Done.')

    def reloadMeetingConfigs(self, full=False):
        '''Reload MeetingConfigs, either only portal_types related parameters,
           or full reload.'''
        logger.info("Reloading every MeetingConfigs (full={0})...".format(repr(full)))
        for cfg in self.tool.objectValues('MeetingConfig'):
            if full:
                notify(ObjectEditedEvent(cfg))
            else:
                cfg.registerPortalTypes()
        logger.info('Done.')

    def updateFacetedFilters(self, xml_filename=None, related_to="items", reorder=True, to_delete=[]):
        """ """
        logger.info("Updating faceted filters for every MeetingConfigs...")

        if xml_filename:
            xmlpath = os.path.join(
                os.path.dirname(__file__),
                '../faceted_conf/%s' % xml_filename)
            logger.info("Applying faceted config at %s..." % xmlpath)

        for cfg in self.tool.objectValues('MeetingConfig'):
            obj = None
            if related_to == "items":
                obj = cfg.searches.searches_items
            elif related_to == "meetings":
                obj = cfg.searches.searches_meetings
            elif related_to == "decisions":
                obj = cfg.searches.searches_decisions
            # add/update faceted filters if xml_filename
            if xml_filename:
                obj.unrestrictedTraverse('@@faceted_exportimport').import_xml(
                    import_file=open(xmlpath))
            # delete criteria id given in to_delete
            cleanRamCacheFor('Products.PloneMeeting.adapters.compute_criteria')
            criteria = ICriteria(obj)
            for cid in to_delete:
                if cid in criteria.keys():
                    logger.info("Removed criterion {0} from {1}".format(
                        cid, repr(obj)))
                    criteria.delete(cid)
            # when criteria have been imported, if some were purged, we need to reorder
            if reorder:
                # sort by criterion name, so c0, c1, c2, ...
                criteria._update(humansorted(criteria.values(), key=attrgetter('__name__')))
        logger.info('Done.')

    def changeCollectionIndex(self, old_index_name, new_index_name):
        """Useful when an index name changed."""
        for cfg in self.tool.objectValues('MeetingConfig'):
            for brain in api.content.find(context=cfg.searches, portal_type='DashboardCollection'):
                dc = brain.getObject()
                query = dc.query
                found = False
                adapted_query = []
                for line in query:
                    adapted_line = line.copy()
                    if adapted_line['i'] == old_index_name:
                        found = True
                        adapted_line['i'] = new_index_name
                        logger.info("Replaced \"{0}\" by \"{1}\" in query of \"{2}\"".format(
                            old_index_name, new_index_name, brain.getPath()))
                    adapted_query.append(adapted_line)
                if base_hasattr(dc, 'sort_on') and dc.sort_on == old_index_name:
                    dc.sort_on = new_index_name
                    logger.info("Replaced \"{0}\" by \"{1}\" in sort_on of \"{2}\"".format(
                        old_index_name, new_index_name, brain.getPath()))
                if found:
                    dc.query = adapted_query
                    dc._p_changed = True
        logger.info('Done.')

    def _removeBrokenAnnexes(self):
        """Remove annexes that do not have a content_category,
           that could happen with quickupload."""
        logger.info("Removing broken annexes, annexes uploaded withtout a content_category...")
        brains = self.catalog(portal_type=['annex', 'annexDecision'])
        pghandler = ZLogHandler(steps=1000)
        pghandler.init('Removing broken annexes', len(brains))
        i = found = 0
        for brain in brains:
            pghandler.report(i)
            annex = brain.getObject()
            parent = annex.aq_inner.aq_parent
            categorized_elements = _categorized_elements(parent)
            if annex.UID() not in categorized_elements:
                logger.info('In _removeBrokenAnnexes, removed %s' % brain.getPath())
                # make sure parent is not modified
                parent_modified = parent.modified()
                parent.manage_delObjects(ids=[annex.getId()])
                parent.setModificationDate(parent_modified)
                parent.reindexObject(idxs=['modified', 'ModificationDate', 'Date'])
                found += 1
            i += 1
        if found:
            self.warn(logger, 'In _removeBrokenAnnexes, removed %s annexe(s)' % found)
        logger.info('Done.')

    def addCKEditorStyle(self, style_name, style_element, style_type="class", style_value=None):
        """Helper for adding a new style to the CKeditor styles.
           The rules:
           - p_style_name may use "-" or "_";
           - p_style_element is the tag name, so "p" or "span";
           - the corresponding translation is "ckeditor_style_" +
             style_name where "-" are replaced by "_";
           - p_style_type is the type of style, by default "class", could be "style" also;
           - p_style_value is the style value (class name or style definition)."""
        logger.info("Adding style '%s' to CKEditor styles..." % style_name)
        cke_props = self.portal.portal_properties.ckeditor_properties
        if cke_props.menuStyles.find(style_name) == -1:
            # msgid always starts with ckeditor_style_ and ends with style_name
            msg_style_name = translate(
                'ckeditor_style_{0}'.format(style_name),
                domain='PloneMeeting',
                context=self.request)
            menuStyles = cke_props.menuStyles
            style = u"{{ name : '{0}'\t\t, element : '{1}', attributes : " \
                u"{{ '{2}' : '{3}' }} }},\n]".format(
                    msg_style_name, style_element, style_type, style_value)
            # last element, check if we need a ',' before or not...
            strippedMenuStyles = menuStyles.replace(u' ', u'').replace(u'\n', u'').replace(u'\r', u'')
            if u',]' not in strippedMenuStyles:
                menuStyles = menuStyles.replace(u'\n]', u']')
                style = u",\n" + style
            menuStyles = menuStyles.replace(u']', style)
            cke_props.menuStyles = menuStyles
            self.warn(logger, "Style '{0}' was added...".format(style_name))
        else:
            logger.info("Style '{0}' already exists and was not added...".format(style_name))
        logger.info('Done.')

    def _already_migrated(self, done=True):
        """Called when a migration is executed several times..."""
        self.already_migrated = True
        logger.info('Already migrated ...')
        if done:
            logger.info('Done.')

    def _warnPortalSkinsCustom(self):
        """Add a warning for each value found in portal_skins/custom."""
        values = self.portal.portal_skins.custom.objectIds()
        for value in values:
            self.warn(logger, "\"%s\" was found in portal_skins/custom, still necessary?" % value)

    def run(self):
        '''Must be overridden. This method does the migration job.'''
        raise 'You should have overridden me darling.'''

    def finish(self):
        '''At the end of the migration, you can call this method to log its
           duration in minutes.'''
        # set mailMode for every MeetingConfigs back to the right value
        for cfgId in self.cfgsMailMode:
            cfg = getattr(self.tool, cfgId)
            cfg.setMailMode(self.cfgsMailMode[cfgId])
        # set adviceInvalidation for every MeetingConfigs back to the right value
        for cfgId in self.cfgsAdvicesInvalidation:
            cfg = getattr(self.tool, cfgId)
            cfg.setEnableAdviceInvalidation(self.cfgsAdvicesInvalidation[cfgId])
        self._warnPortalSkinsCustom()
        self.cleanRegistries()
        self.tool.invalidateAllCache()
        BaseMigrator.finish(self)
        logger.info('======================================================================')



# encoding: utf-8

from ftw.labels.interfaces import ILabelJar
from imio.helpers import EMPTY_STRING
from imio.helpers.cache import get_cachekey_volatile
from imio.helpers.cache import get_current_user_id
from plone import api
from plone.memoize import ram
from Products.PloneMeeting.ftw_labels.utils import filter_access_global_labels
from Products.PloneMeeting.utils import get_context_with_request
from zope.i18n import translate
from zope.interface import implements
from zope.schema.interfaces import IVocabularyFactory
from zope.schema.vocabulary import SimpleTerm
from zope.schema.vocabulary import SimpleVocabulary


class FTWLabelsVocabulary(object):
    """
        Vocabulary that lists available ftw.labels
        labels for the current MeetingConfig.
        Used for DashboardCollection query.
    """
    implements(IVocabularyFactory)

    def __call__(self, context, include_personal_labels=True):
        res = []
        context = get_context_with_request(context)

        tool = api.portal.get_tool('portal_plonemeeting')
        try:
            # in some case, like Plone Site creation, context is the Zope app...
            cfg = tool.getMeetingConfig(context)
        except Exception:
            return SimpleVocabulary(res)

        if cfg and 'labels' in cfg.getUsedItemAttributes():
            labels = ILabelJar(cfg).list()
            for label in labels:
                if label['by_user']:
                    if include_personal_labels is False:
                        continue
                    title = '{0} ({1}) (*)'.format(label['title'], label['label_id'])
                else:
                    title = '{0} ({1})'.format(label['title'], label['label_id'])
                res.append(SimpleTerm(
                    label['label_id'],
                    label['label_id'],
                    title))

        return SimpleVocabulary(res)


FTWLabelsVocabularyFactory = FTWLabelsVocabulary()


class ConfigFTWLabelsVocabulary(FTWLabelsVocabulary):
    """
        Vocabulary used for MeetingConfig.labelsConfig
    """
    def __call__(self, context):
        res = super(ConfigFTWLabelsVocabulary, self).__call__(
            context, include_personal_labels=False)
        res._terms.insert(
            0,
            SimpleTerm("*",
                       "*",
                       '{0} (*)'.format(
                        translate('default_for_all_labels',
                                  domain='PloneMeeting',
                                  context=context.REQUEST).encode('utf-8'))))
        return res


ConfigFTWLabelsVocabularyFactory = ConfigFTWLabelsVocabulary()


class FTWLabelsForFacetedFilterVocabulary(object):
    """
        Vocabulary that lists available ftw.labels
        labels for the current MeetingConfig.
        Use in faceted navigation, is current user aware.
    """
    implements(IVocabularyFactory)

    def __call___cachekey(method, self, context):
        '''cachekey method for self.__call__.'''
        date = get_cachekey_volatile(
            'Products.PloneMeeting.vocabularies.ftwlabelsforfacetedfiltervocabulary')
        tool = api.portal.get_tool('portal_plonemeeting')
        cfg = tool.getMeetingConfig(context)
        # personal labels include current user id, more over global labels
        # are not viewable by everybody
        return date, cfg.getId(), cfg.modified(), get_current_user_id(context.REQUEST)

    @ram.cache(__call___cachekey)
    def __call__(self, context):
        context = get_context_with_request(context)
        tool = api.portal.get_tool('portal_plonemeeting')
        cfg = tool.getMeetingConfig(context)
        member_id = get_current_user_id(context.REQUEST)

        res = []
        # ftw.labels will index "_" when no label selected
        res.append(
            SimpleTerm(EMPTY_STRING,
                       EMPTY_STRING,
                       translate('(None)',
                                 domain='PloneMeeting',
                                 context=context.REQUEST)))

        # only keep labels the user can view
        labels = filter_access_global_labels(ILabelJar(cfg))
        for label in labels:
            if label['by_user']:
                res.append(SimpleTerm(
                    '{0}:{1}'.format(member_id, label['label_id']),
                    '{0}:{1}'.format(member_id, label['label_id']),
                    '{0} (*)'.format(label['title'])))
            else:
                res.append(SimpleTerm(
                    label['label_id'],
                    label['label_id'],
                    label['title']))
        return SimpleVocabulary(res)


FTWLabelsForFacetedFilterVocabularyFactory = FTWLabelsForFacetedFilterVocabulary()



# -*- coding: utf-8 -*-

from collective.contact.core.content.organization import IOrganization
from collective.contact.core.content.organization import Organization
from collective.contact.plonegroup.config import PLONEGROUP_ORG
from collective.contact.plonegroup.interfaces import IPloneGroupContact
from collective.contact.plonegroup.utils import get_organizations
from collective.z3cform.datagridfield import DataGridFieldFactory
from collective.z3cform.datagridfield import DictRow
from datetime import date
from datetime import datetime
from DateTime import DateTime
from imio.helpers.content import get_back_relations
from imio.helpers.content import uuidsToObjects
from plone.autoform import directives as form
from plone.dexterity.schema import DexteritySchemaPolicy
from plone.supermodel import model
from Products.PloneMeeting.config import PMMessageFactory as _
from Products.PloneMeeting.events import _invalidateOrgRelatedCachedVocabularies
from Products.PloneMeeting.utils import computeCertifiedSignatures
from Products.PloneMeeting.utils import listifySignatures
from Products.PloneMeeting.validators import DXCertifiedSignaturesValidator
from Products.PloneMeeting.widgets.pm_checkbox import PMCheckBoxFieldWidget
from z3c.form import validator
from zope import schema
from zope.interface import Interface


class ICertifiedSignaturesRowSchema(Interface):
    """Schema for DataGridField widget's row of field 'certified_signatures'."""

    signature_number = schema.Choice(
        title=_("Certified signatures signature number"),
        description=_("Select the signature number, keep signatures ordered by number."),
        vocabulary='Products.PloneMeeting.vocabularies.numbersvocabulary',
        required=True,
    )

    name = schema.TextLine(
        title=_(u'Certified signatures signatory name'),
        description=_("Name of the signatory (for example 'Mister John Doe')."),
        required=False,
    )

    function = schema.TextLine(
        title=_(u"Certified signatures signatory function"),
        description=_("Function of the signatory (for example 'Mayor')."),
        required=False,
    )

    held_position = schema.Choice(
        title=_("Certified signatures held position"),
        description=_("Select a held position if necessary, 'Name', 'Function' and other data of this "
                      "held position will be used if you leave 'Name' and 'Function' columns empty."),
        vocabulary='Products.PloneMeeting.vocabularies.every_heldpositions_vocabulary',
        required=False,
    )

    date_from = schema.Date(
        title=_("Certified signatures valid from (included)"),
        description=_("Enter valid from date, use following format : YYYY/MM/DD, "
                      "leave empty so it is always valid."),
        required=False,
        default=None,
    )

    date_to = schema.Date(
        title=_("Certified signatures valid to (included)"),
        description=_("Enter valid to date, use following format : YYYY/MM/DD, "
                      "leave empty so it is always valid."),
        required=False,
        default=None,
    )


class IPMOrganization(IOrganization):
    """These fields are for organizations added to the 'plonegroup-organization' organization.
       We protect these fields with read/write permission so it is only
       shown on organization added to 'plonegroup-organization' and fields are hidden
       from view and edit form when adding an organization outside "My organization"."""

    acronym = schema.TextLine(
        title=_("PloneMeeting_label_acronym"),
        required=False,
    )

    form.read_permission(item_advice_states='PloneMeeting.manage_internal_organization_fields')
    form.write_permission(item_advice_states='PloneMeeting.manage_internal_organization_fields')
    form.widget('item_advice_states', PMCheckBoxFieldWidget, multiple='multiple')
    item_advice_states = schema.List(
        title=_("PloneMeeting_label_itemAdviceStates"),
        description=_("group_item_advice_states_descr"),
        value_type=schema.Choice(
            vocabulary="Products.PloneMeeting.vocabularies.itemallstates"),
        required=False,
        default=[],
    )

    form.read_permission(item_advice_edit_states='PloneMeeting.manage_internal_organization_fields')
    form.write_permission(item_advice_edit_states='PloneMeeting.manage_internal_organization_fields')
    form.widget('item_advice_edit_states', PMCheckBoxFieldWidget, multiple='multiple')
    item_advice_edit_states = schema.List(
        title=_("PloneMeeting_label_itemAdviceEditStates"),
        description=_("group_item_advice_edit_states_descr"),
        value_type=schema.Choice(
            vocabulary="Products.PloneMeeting.vocabularies.itemallstates"),
        required=False,
        default=[],
    )

    form.read_permission(item_advice_view_states='PloneMeeting.manage_internal_organization_fields')
    form.write_permission(item_advice_view_states='PloneMeeting.manage_internal_organization_fields')
    form.widget('item_advice_view_states', PMCheckBoxFieldWidget, multiple='multiple')
    item_advice_view_states = schema.List(
        title=_("PloneMeeting_label_itemAdviceViewStates"),
        description=_("group_item_advice_view_states_descr"),
        value_type=schema.Choice(
            vocabulary="Products.PloneMeeting.vocabularies.itemallstates"),
        required=False,
        default=[],
    )

    form.read_permission(keep_access_to_item_when_advice='PloneMeeting.manage_internal_organization_fields')
    form.write_permission(keep_access_to_item_when_advice='PloneMeeting.manage_internal_organization_fields')
    keep_access_to_item_when_advice = schema.Choice(
        title=_(u'PloneMeeting_label_keepAccessToItemWhenAdvice'),
        description=_("keep_access_to_item_when_advice_descr"),
        vocabulary=u'Products.PloneMeeting.vocabularies.keep_access_to_item_when_advice_vocabulary',
        default='use_meetingconfig_value',
        required=True,
    )

    form.read_permission(as_copy_group_on='PloneMeeting.manage_internal_organization_fields')
    form.write_permission(as_copy_group_on='PloneMeeting.manage_internal_organization_fields')
    as_copy_group_on = schema.TextLine(
        title=_("PloneMeeting_label_asCopyGroupOn"),
        description=_("as_copy_group_on_descr"),
        required=False,
    )

    form.read_permission(as_restricted_copy_group_on='PloneMeeting.manage_internal_organization_fields')
    form.write_permission(as_copy_group_on='PloneMeeting.manage_internal_organization_fields')
    as_restricted_copy_group_on = schema.TextLine(
        title=_("PloneMeeting_label_asRestrictedCopyGroupOn"),
        description=_("as_restricted_copy_group_on_descr"),
        required=False,
    )

    form.read_permission(certified_signatures='PloneMeeting.manage_internal_organization_fields')
    form.write_permission(certified_signatures='PloneMeeting.manage_internal_organization_fields')
    form.widget('certified_signatures', DataGridFieldFactory, allow_reorder=True)
    certified_signatures = schema.List(
        title=_(u'PloneMeeting_label_group_certifiedSignatures'),
        description=_("group_certified_signatures_descr"),
        required=False,
        value_type=DictRow(
            schema=ICertifiedSignaturesRowSchema,
            required=False
        ),
        default=[],
    )

    form.read_permission(groups_in_charge='PloneMeeting.manage_internal_organization_fields')
    form.write_permission(groups_in_charge='PloneMeeting.manage_internal_organization_fields')
    form.widget('groups_in_charge', PMCheckBoxFieldWidget, multiple='multiple')
    groups_in_charge = schema.List(
        title=_("PloneMeeting_label_groupsInCharge"),
        description=_("groups_in_charge_descr"),
        value_type=schema.Choice(
            vocabulary="collective.contact.plonegroup.browser.settings."
            "SortedSelectedOrganizationsElephantVocabulary"),
        required=False,
        default=[],
    )

    model.fieldset('app_parameters',
                   label=_(u"Application parameters"),
                   fields=['acronym', 'item_advice_states',
                           'item_advice_edit_states', 'item_advice_view_states',
                           'keep_access_to_item_when_advice',
                           'as_copy_group_on', 'as_restricted_copy_group_on',
                           'certified_signatures', 'groups_in_charge'])


validator.WidgetValidatorDiscriminators(
    DXCertifiedSignaturesValidator,
    field=IPMOrganization['certified_signatures']
)


class PMOrganization(Organization):
    """ """

    def get_acronym(self):
        """Accessor so it can be called in a TAL expression."""
        return self.acronym or u''

    def get_groups_in_charge(self, the_objects=False):
        """Accessor so it can be called in a TAL expression."""
        res = self.groups_in_charge
        if res and the_objects:
            res = uuidsToObjects(res, ordered=True, unrestricted=True)
        return res

    def get_full_title_cachekey(method, self, separator=u' / ', first_index=0, force_separator=False):
        '''cachekey method for self.get_full_title.'''
        return repr(self), self.modified(), separator, first_index, force_separator

    # not ramcached perf tests says it does not change much
    # and this avoid useless entry in cache
    # @ram.cache(get_full_title_cachekey)
    def get_full_title(self, separator=u' / ', first_index=0, force_separator=False):
        """Override to change default first_index from 0 to 1 for IPloneGroupContact,
           so we do not display the 'My organization' level for elements displayed in the
           'My organization' organization."""
        if not force_separator and self.id != PLONEGROUP_ORG and IPloneGroupContact.providedBy(self):
            first_index = 1
        return super(PMOrganization, self).get_full_title(separator, first_index)

    def get_item_advice_states(self, cfg=None):
        res = self.item_advice_states
        if cfg:
            tmpres = []
            givenCfgId = cfg.getId()
            for elt in res:
                cfgId, state = elt.split('__state__')
                if cfgId == givenCfgId:
                    tmpres.append(state)
            # if nothing redefined for given p_cfg in this organization,
            # use value defined on the cfg
            res = tmpres or cfg.getItemAdviceStates()
        return res

    def get_item_advice_edit_states(self, cfg=None):
        res = self.item_advice_edit_states
        if cfg:
            tmpres = []
            givenCfgId = cfg.getId()
            for elt in res:
                cfgId, state = elt.split('__state__')
                if cfgId == givenCfgId:
                    tmpres.append(state)
            # if nothing redefined for given p_cfg in this organization,
            # use value defined on the cfg
            res = tmpres or cfg.getItemAdviceEditStates()
        return res

    def get_item_advice_view_states(self, cfg=None):
        res = self.item_advice_view_states
        if cfg:
            tmpres = []
            givenCfgId = cfg.getId()
            for elt in res:
                cfgId, state = elt.split('__state__')
                if cfgId == givenCfgId:
                    tmpres.append(state)
            # if nothing redefined for given p_cfg in this organization,
            # use value defined on the cfg
            res = tmpres or cfg.getItemAdviceViewStates()
        return res

    def get_keep_access_to_item_when_advice(self, cfg=None):
        """ """
        res = self.keep_access_to_item_when_advice
        if cfg and res == 'use_meetingconfig_value':
            res = cfg.getKeepAccessToItemWhenAdvice()
        return res

    def get_certified_signatures(self, computed=False, cfg=None, group_in_charge=None, listify=True, **kwargs):
        """Overrides field 'certified_signatures' accessor to be able to pass
           the p_computed parameter that will return computed certified signatures,
           so signatures really available right now.  If nothing is defined on the organization,
           use certified signatures defined on the corresponding p_cfg MeetingConfig.
           If p_from_group_in_charge is an organization, we get certifiedSignatures from it."""
        group_signatures = self.certified_signatures
        if computed:
            computed_signatures = cfg.getCertifiedSignatures(computed=True)

            # get certified signatures from first of the defined groupsInCharge
            if group_in_charge:
                computed_signatures.update(computeCertifiedSignatures(group_in_charge.get_certified_signatures()))

            # if we have certified signatures defined on this MeetingGroup
            # update MeetingConfig signatures regarding what is defined here
            if group_signatures:
                computed_signatures.update(computeCertifiedSignatures(group_signatures))
            # listify signatures, for backward compatibility, we need a list of pair
            # of function/name, like ['function1', 'name1', 'function2', 'name2']
            if listify:
                group_signatures = listifySignatures(computed_signatures)
            else:
                group_signatures = computed_signatures
        return group_signatures

    def get_order(self, associated_org_uids=[], cfg=None):
        '''Returns organization position among every selected organizations.
           If p_associated_org_uids is given, returns the order of the lowest org position.
           In this case, p_cfg must be given.'''
        def _get_index(orgs, org):
            """Return position of org among orgs, return 0 if org not found (not selected),
               it it like if it was using the first organization."""
            try:
                # +1 to index as index 0 is for no more selected organizations
                index = orgs.index(org) + 1
            except ValueError:
                index = 0
            return index
        org_uids = get_organizations(only_selected=True, the_objects=False)
        i = _get_index(org_uids, self.UID())
        # if we received associated_org_uids we must consider associated group
        # that has the lowest position
        if associated_org_uids:
            # if we have MeetingConfig.orderedAssociatedOrganizations, we use it
            # either we use organizations selected in plonegroup
            org_uids = cfg.getOrderedAssociatedOrganizations() or org_uids
            # orgs are sorted so, the first we find, we return it
            for org_uid in org_uids:
                if org_uid in associated_org_uids:
                    # we found the associated org with lowest position, now check
                    # that the lowest position of this associated group is lower or not
                    # than the position of the proposing group
                    associated_org_index = _get_index(org_uids, org_uid)
                    if associated_org_index < i:
                        i = associated_org_index
                    break
        return i

    def get_representatives(self, at_date=None):
        '''Get the representative held positions.
           When a date is given, it will get the held position that were active at given date.
           Else, date will be set to "now" to only get the currently active held positions.'''
        if not at_date:
            at_date = date.today()
        elif isinstance(at_date, DateTime):
            at_date = at_date.asdatetime().date()
        elif isinstance(at_date, datetime):
            at_date = at_date.date()
        res = []
        for hp in get_back_relations(self, 'represented_organizations'):
            if not hp.end_date or hp.end_date >= at_date:
                res.append(hp)
        return res

    def _invalidateCachedMethods(self):
        '''Clean cache for various methods using organizations.'''
        _invalidateOrgRelatedCachedVocabularies()


class PMOrganizationSchemaPolicy(DexteritySchemaPolicy):
    """ """

    def bases(self, schemaName, tree):
        return (IPMOrganization, )
